# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import warnings
from typing import Optional

import numpy as np
import pandas as pd

from ..base import Module
from ..utils import parallel


class ApplyFunc(Module):
    """This module applies functions to specified feature and metrics.

    Extra parameters (kwargs) can be passed to the apply function.
    """

    _input_keys = ("apply_to_key", "assign_to_key", "apply_funcs_key")
    _output_keys = ("store_key",)

    def __init__(
        self,
        apply_to_key,
        store_key="",
        assign_to_key="",
        apply_funcs_key="",
        features=None,
        apply_funcs=None,
        metrics=None,
        msg="",
    ):
        """Initialize an instance of ApplyFunc.

        :param str apply_to_key: key of the input data to apply funcs to.
        :param str assign_to_key: key of the input data to assign function applied-output to. (optional)
        :param str store_key: key of the output data to store in the datastore (optional)
        :param str apply_funcs_key: key of to-be-applied functions in data to store (optional)
        :param list features: list of features to pick up from input data and apply funcs to (optional)
        :param list metrics: list of metrics to apply funcs to (optional)
        :param str msg: message to print out at start of transform function. (optional)
        :param list apply_funcs: functions to apply (list of dicts):

          - 'func': function to apply
          - 'suffix' (string, optional): suffix added to each metric. default is function name.
          - 'prefix' (string, optional): prefix added to each metric.
          - 'features' (list, optional): features the function is applied to. Overwrites features above
          - 'metrics' (list, optional): metrics the function is applied to. Overwrites metrics above
          - 'entire' (boolean, optional): apply function to the entire feature's dataframe of metrics?
          - 'args' (tuple, optional): args for 'func'
          - 'kwargs' (dict, optional): kwargs for 'func'

        """
        super().__init__()
        self.apply_to_key = apply_to_key
        self.assign_to_key = assign_to_key or apply_to_key
        self.apply_funcs_key = apply_funcs_key
        self.store_key = store_key or self.assign_to_key

        self.features = features or []
        self.metrics = metrics or []
        self.msg = msg
        self.apply_funcs = []
        # import applied functions
        apply_funcs = apply_funcs or []
        for af in apply_funcs:
            self.add_apply_func(**af)

    def get_description(self):
        if len(self.apply_funcs) > 0:
            return " and ".join([x["func"].__name__ for x in self.apply_funcs])
        elif self.apply_funcs_key:
            return f"functions from arg '{self.apply_funcs_key}'"
        else:
            raise NotImplementedError

    def add_apply_func(
        self,
        func,
        suffix=None,
        prefix=None,
        metrics=[],
        features=[],
        entire=None,
        *args,
        **kwargs,
    ):
        """Add function to be applied to dataframe.

        Can call this function after module instantiation to add new functions.

        :param func: function to apply
        :param suffix: (string, optional) suffix added to each metric. default is function name.
        :param prefix: (string, optional) prefix added to each metric.
        :param features: (list, optional) features the function is applied to. Overwrites features above
        :param metrics: (list, optional) metrics the function is applied to. Overwrites metrics above
        :param entire: (boolean, optional) apply function to the entire feature's dataframe of metrics?
        :param args: (tuple, optional) args for 'func'
        :param kwargs: (dict, optional) kwargs for 'func'
        """
        # check inputs
        if not callable(func):
            raise TypeError("functions in ApplyFunc must be callable objects")
        if suffix is not None and not isinstance(suffix, str):
            raise TypeError("prefix, and suffix in ApplyFunc must be strings or None.")
        if prefix is not None and not isinstance(prefix, str):
            raise TypeError("prefix, and suffix in ApplyFunc must be strings or None.")
        if not isinstance(metrics, list) or not isinstance(features, list):
            raise TypeError("metrics and features must be lists of strings.")

        # add function
        self.apply_funcs.append(
            {
                "features": features,
                "metrics": metrics,
                "func": func,
                "entire": entire,
                "suffix": suffix,
                "prefix": prefix,
                "args": args,
                "kwargs": kwargs,
            }
        )

    def transform(
        self,
        apply_to_data: dict,
        assign_to_data: Optional[dict] = None,
        apply_funcs: Optional[list] = None,
    ):
        """
        Apply functions to specified feature and metrics

        Each feature/metric combination is treated as a pandas series

        :param datastore: input datastore
        :return: updated datastore
        :rtype: dict
        """
        assert isinstance(apply_to_data, dict)
        if assign_to_data is None:
            assign_to_data = {}

        if apply_funcs is not None:
            self.apply_funcs += apply_funcs

        if self.msg:
            self.logger.info(self.msg)

        features = self.get_features(list(apply_to_data.keys()))

        same_key = self.assign_to_key == self.apply_to_key

        args = [
            {
                "feature": feature,
                "metrics": self.metrics,
                "apply_to_df": self.get_datastore_object(
                    apply_to_data, feature, dtype=pd.DataFrame
                ),
                "assign_to_df": (
                    None
                    if same_key
                    else self.get_datastore_object(
                        assign_to_data,
                        feature,
                        dtype=pd.DataFrame,
                        default=pd.DataFrame(),
                    )
                ),
                "apply_funcs": self.apply_funcs,
                "same_key": same_key,
            }
            for feature in features
        ]
        result = parallel(apply_func_array, args, mode="kwargs")
        new_metrics = dict(result)
        return new_metrics


def apply_func_array(
    feature, metrics, apply_to_df, assign_to_df, apply_funcs, same_key
):
    """Apply list of functions to dataframe

    Split off for parallelization reasons

    :param str feature: feature currently looping over
    :param list metrics: list of selected metrics to apply functions to
    :param apply_to_df: pandas data frame that function in arr is applied to
    :param assign_to_df: pandas data frame the output of function is assigned to
    :param apply_funcs: list of functions to apply to
    :param same_key: if True, merge apply_to_df and assign_to_df before returning assign_to_df
    :return: union of feature and assign_to_df
    """
    if not isinstance(apply_to_df, pd.DataFrame):
        raise TypeError(
            f'apply_to_df of feature "{feature}" is not a pandas dataframe.'
        )

    if same_key or (len(assign_to_df.index) == 0 and len(assign_to_df.columns) == 0):
        assign_to_df = pd.DataFrame(index=apply_to_df.index)

    for arr in apply_funcs:
        obj = apply_func(feature, metrics, apply_to_df, arr)
        if len(obj) == 0:
            # no metrics were found in apply_to_df
            continue
        for new_metric, o in obj.items():
            if isinstance(o, pd.Series):
                if len(assign_to_df.index) == len(o) and all(
                    assign_to_df.index == o.index
                ):
                    assign_to_df[new_metric] = o
                else:
                    warnings.warn(
                        f"{feature}:{new_metric}: df_out and object have inconsistent lengths."
                    )
            else:
                # o is number or object, assign to every element of new column
                assign_to_df[new_metric] = [o] * len(assign_to_df.index)
    if same_key:
        assign_to_df = pd.concat([apply_to_df, assign_to_df], axis=1)
    return feature, assign_to_df


def apply_func(feature, selected_metrics, df, arr):
    """Apply function to dataframe

    :param str feature: feature currently looping over
    :param list selected_metrics: list of selected metrics to apply to
    :param df: pandas data frame that function in arr is applied to
    :param dict arr: dictionary containing the function to be applied to pandas dataframe.
    :return: dictionary with outputs of applied-to metric pd.Series
    """
    # basic checks of feature
    if "features" in arr and len(arr["features"]) > 0:
        if feature not in arr["features"]:
            return {}

    # get func input
    keys = list(arr.keys())

    assert "func" in keys, "function input is insufficient."
    func = arr["func"]

    if "prefix" not in keys or arr["prefix"] is None:
        arr["prefix"] = ""
    if len(arr["prefix"]) > 0 and not arr["prefix"].endswith("_"):
        arr["prefix"] = arr["prefix"] + "_"
    if "suffix" not in keys or arr["suffix"] is None:
        arr["suffix"] = func.__name__ if len(arr["prefix"]) == 0 else ""
    if len(arr["suffix"]) > 0 and not arr["suffix"].startswith("_"):
        arr["suffix"] = "_" + arr["suffix"]
    suffix = arr["suffix"]
    prefix = arr["prefix"]

    args = ()
    kwargs = {}
    if "kwargs" in keys:
        kwargs = arr["kwargs"]
    if "args" in keys:
        args = arr["args"]

    # apply func
    if len(selected_metrics) > 0 or ("metrics" in keys and len(arr["metrics"]) > 0):
        metrics = (
            arr["metrics"]
            if ("metrics" in keys and len(arr["metrics"]) > 0)
            else selected_metrics
        )
        metrics = [m for m in metrics if m in df.columns]
        if len(metrics) == 0:
            return {}
        df = df[metrics] if len(metrics) >= 2 else df[metrics[0]]

    if (
        "entire" in arr
        and arr["entire"] is not None
        and arr["entire"] is not False
        and arr["entire"] != 0
    ):
        obj = func(df, *args, **kwargs)
    else:
        obj = df.apply(func, args=args, **kwargs)

    # convert object to dict format
    if not isinstance(
        obj, (pd.Series, pd.DataFrame, list, tuple, np.ndarray)
    ) and isinstance(df, pd.Series):
        obj = {df.name: obj}
    elif not isinstance(
        obj, (pd.Series, pd.DataFrame, list, tuple, np.ndarray)
    ) and isinstance(df, pd.DataFrame):
        obj = {"_".join(df.columns): obj}
    elif (
        isinstance(obj, (list, tuple, np.ndarray))
        and isinstance(df, pd.DataFrame)
        and len(df.columns) == len(obj)
    ):
        obj = {c: o for c, o in zip(df.columns, obj)}
    elif (
        isinstance(obj, (list, tuple, np.ndarray))
        and isinstance(df, pd.Series)
        and len(df.index) == len(obj)
    ):
        obj = {df.name: pd.Series(data=obj, index=df.index)}
    elif (
        isinstance(obj, (list, tuple, np.ndarray))
        and isinstance(df, pd.DataFrame)
        and len(df.index) == len(obj)
    ):
        obj = {"_".join(df.columns): pd.Series(data=obj, index=df.index)}
    elif (
        isinstance(obj, pd.Series)
        and isinstance(df, pd.Series)
        and len(obj) == len(df)
        and all(obj.index == df.index)
    ):
        obj = {df.name: obj}
    elif (
        isinstance(obj, pd.Series)
        and isinstance(df, pd.DataFrame)
        and len(obj) == len(df)
        and all(obj.index == df.index)
    ):
        obj = {"_".join(df.columns): obj}
    elif (
        isinstance(obj, pd.DataFrame)
        and len(obj.columns) == 1
        and len(obj.index) != len(df.index)
    ):
        # e.g. output of normalized_hist_mean_cov: a dataframe with one column, actually a series
        obj = obj[obj.columns[0]].to_dict()
    elif (
        isinstance(obj, pd.DataFrame)
        and len(obj.columns) == 1
        and len(obj.index) == len(df.index)
        and (obj.index != df.index).any()
    ):
        # e.g. output of normalized_hist_mean_cov: a dataframe with one column, actually a series
        obj = obj[obj.columns[0]].to_dict()
    elif isinstance(obj, pd.Series):
        # e.g. output of np.mean of np.std: results in one number per column when applied to a dataframe
        obj = obj.to_dict()
    elif isinstance(obj, pd.DataFrame):
        # when apply to one column returns another column
        obj = {c: obj[c] for c in obj.columns}
    assert isinstance(obj, dict)

    # add prefix and suffix to dict keys
    keys = list(obj.keys())
    for k in keys:
        obj[prefix + k + suffix] = obj.pop(k)

    return obj



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from math import floor
from typing import List, Optional, Union

import pandas as pd
from typing_extensions import Literal


def _split_dataset_pandas(
    dataset,
    split_type: Literal["n_instances", "fraction", "condition"],
    split: Union[int, float, pd.Series, str],
):
    if split_type in ["n_instances", "fraction"]:
        if split_type == "fraction":
            n = dataset.shape[0]
            split = floor(split * n)
        reference = dataset.iloc[:split]
        df = dataset.iloc[split:]
    else:
        reference = dataset[split]
        df = dataset[~split]

    if reference.empty:
        raise ValueError("Reference is empty. Please adjust the `split` argument")
    if df.empty:
        raise ValueError(
            "Returned dataframe is empty. Please adjust the `split` argument"
        )

    return reference, df


def _split_dataset_spark(
    dataset,
    split_type: Literal["n_instances", "fraction", "condition"],
    split: Union[int, float, pd.Series, str],
    time_axis: str,
    partition_cols: Optional[List[str]] = None,
    persist: bool = True,
):
    """
    Split a dataset into a reference and remaining part based on split params using PySpark
    (see `split_dataset`)

    :param split_type: meaning of split parameter
    :param split: split parameter
    :param time_axis: time axis
    :param partition_cols: cols to partition by (for performance)
    :param persist: persist or not, enabled by default since we are checking for empty dataframes
    """
    from pyspark.sql import functions as F
    from pyspark.sql.window import Window

    if split_type in ["n_instances", "fraction"]:
        data_win = Window.orderBy(time_axis)
        if partition_cols is not None:
            data_win = data_win.partitionBy(partition_cols)

        if split_type == "n_instances":
            dt_rank = dataset.withColumn("tmpcol", F.row_number().over(data_win))
        else:
            dt_rank = dataset.withColumn("tmpcol", F.percent_rank().over(data_win))

        reference = dt_rank.filter(F.col("tmpcol") <= split).drop("tmpcol")
        df = dt_rank.filter(F.col("tmpcol") > split).drop("tmpcol")
    else:
        reference = dataset.filter(split)
        df = dataset.filter(f"NOT ({split})")

    if persist:
        reference.persist()
        df.persist()

    if len(reference.head(1)) == 0:
        raise ValueError("Reference is empty. Please adjust the `split` argument")
    if len(df.head(1)) == 0:
        raise ValueError(
            "Returned dataframe is empty. Please adjust the `split` argument"
        )

    return reference, df


def split_dataset(dataset, split: Union[int, float, pd.Series, str], time_axis: str):
    """Split a dataset into a reference and remaining part based on split params.

    :param pd.Dataset|pyspark.sql.Dataset dataset: dataset as input
    :param Any split: split details, meaning depends on the type:
        if integer, then the reference will be the first ``split`` instances
        if float, then ``split`` will be used as ration (e.g. 0.5 returns a 50/50 split)
        otherwise, the ``split`` are interpreted as condition, where the records for which the condition is
        true are considered the reference, and the other records the remaining dataset.
    :param time_axis: the time axis
    :return: tuple of reference, dataset
    """

    if isinstance(split, int):
        split_type = "n_instances"
        if split <= 0:
            raise ValueError("Number of instances should be greater than 0")
    elif isinstance(split, float):
        if split >= 1.0 or split <= 0:
            raise ValueError("Fraction should be 0 > fraction > 1")
        split_type = "fraction"
    elif isinstance(split, (pd.Series, str)):
        split_type = "condition"
    else:
        raise TypeError("`split` should be int, float, str or pd.Series")

    if isinstance(dataset, pd.DataFrame):
        reference, df = _split_dataset_pandas(dataset, split_type, split)
    else:
        reference, df = _split_dataset_spark(dataset, split_type, split, time_axis)

    return reference, df



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import numpy as np

from popmon.base.registry import Registry

from ...analysis.hist_numpy import get_2dgrid
from ...hist.hist_utils import sum_entries
from ...stats import numpy as pm_np

Profiles = Registry()


@Profiles.register(
    key=["min", "max", "p01", "p05", "p16", "p50", "p84", "p95", "p99"],
    description=[
        "Minimum value",
        "Maximum value",
        "1% percentile",
        "5% percentile",
        "16% percentile",
        "50% percentile (median)",
        "84% percentile",
        "95% percentile",
        "99% percentile",
    ],
    dim=1,
    htype="num",
)
def profile_quantiles(x, w, bin_width):
    return tuple(
        pm_np.quantile(
            x, q=[0.0, 1.0, 0.01, 0.05, 0.16, 0.50, 0.84, 0.95, 0.99], weights=w
        )
    )


@Profiles.register(key="mean", description="Mean value", dim=1, htype="num")
def profile_mean(x, w, bin_width):
    return pm_np.mean(x, w)


@Profiles.register(key="std", description="Standard deviation", dim=1, htype="num")
def profile_std(x, w, bin_width):
    return pm_np.std(x, w)


@Profiles.register(
    key="nan", description="Number of missing entries (NaN)", dim=1, htype=None
)
def profile_nan(hist):
    if hasattr(hist, "nanflow"):
        return hist.nanflow.entries
    elif hasattr(hist, "bins") and "NaN" in hist.bins:
        return hist.bins["NaN"].entries
    return 0


@Profiles.register(
    key="overflow",
    description="Number of values larger than the maximum bin-edge of the histogram.",
    dim=1,
    htype=None,
)
def profile_overflow(hist):
    if hasattr(hist, "overflow"):
        return hist.overflow.entries
    return 0


@Profiles.register(
    key="underflow",
    description="Number of values smaller than the minimum bin-edge of the histogram.",
    dim=1,
    htype=None,
)
def profile_underflow(hist):
    if hasattr(hist, "underflow"):
        return hist.underflow.entries
    return 0


@Profiles.register(
    key="phik",
    description="phi-k correlation between the two variables of the histogram",
    dim=2,
    htype=None,
)
def profile_phik(hist):
    from phik import phik

    # calculate phik correlation
    try:
        grid = get_2dgrid(hist)
    except Exception:
        raise

    try:
        phi_k = phik.phik_from_hist2d(observed=grid)
    except ValueError:
        # self.logger.debug(
        #     f"Not enough values in the 2d `{name}` time-split histogram to apply the phik test."
        # )
        phi_k = np.nan
    return phi_k


@Profiles.register(
    key="count", description="Number of entries (non-NaN and NaN)", dim=-1, htype=None
)
def profile_count(hist):
    return int(sum_entries(hist))


@Profiles.register(
    key="entropy",
    description="Entropy in nats",
    dim=-1,
    htype=None,
)
def profile_entropy(hist):
    h = hist.bin_entries()
    h = h / h.sum()
    return -(h * np.ma.log(h)).sum()


@Profiles.register(
    key="filled",
    description="Number of non-missing entries (non-NaN)",
    dim=1,
    htype="all",
)
def profile_filled(_, bin_counts):
    return bin_counts.sum()


@Profiles.register(
    key="distinct", description="Number of distinct entries", dim=1, htype="all"
)
def profile_distinct(bin_labels, bin_counts):
    return len(np.unique(bin_labels[bin_counts > 0]))


@Profiles.register(
    key="fraction_of_true",
    description="Compute fraction of 'true' (as in boolean) labels",
    dim=1,
    htype="cat",
)
def profile_fraction_of_true(bin_labels, bin_counts):
    """Compute fraction of 'true' labels

    :param bin_labels: Array containing numbers whose mean is desired. If `a` is not an
        array, a conversion is attempted.
    :param bin_entries: Array containing weights for the elements of `a`. If `weights` is not an
        array, a conversion is attempted.
    :return: fraction of 'true' labels
    """
    bin_labels = np.array(bin_labels)
    bin_entries = np.array(bin_counts)
    assert len(bin_labels) == len(bin_entries)

    def replace(bl):
        if bl in {"True", "true"}:
            return True
        elif bl in {"False", "false"}:
            return False
        return np.nan

    # basic checks: dealing with boolean labels
    # also accept strings of 'True' and 'False'
    if len(bin_labels) == 0 or len(bin_labels) > 4 or np.sum(bin_entries) == 0:
        return np.nan
    if not np.all([isinstance(bl, (bool, np.bool_)) for bl in bin_labels]):
        if not np.all(
            [isinstance(bl, (str, np.str_, np.string_)) for bl in bin_labels]
        ):
            return np.nan
        # all strings from hereon
        n_true = (bin_labels == "True").sum() + (bin_labels == "true").sum()
        n_false = (bin_labels == "False").sum() + (bin_labels == "false").sum()
        n_nan = (
            (bin_labels == "NaN").sum()
            + (bin_labels == "nan").sum()
            + (bin_labels == "None").sum()
            + (bin_labels == "none").sum()
            + (bin_labels == "Null").sum()
            + (bin_labels == "null").sum()
        )
        if n_true + n_false + n_nan != len(bin_labels):
            return np.nan
        # convert string to boolean
        bin_labels = np.array([replace(bl) for bl in bin_labels])

    sum_true = np.sum([be for bl, be in zip(bin_labels, bin_entries) if bl])
    sum_false = np.sum([be for bl, be in zip(bin_labels, bin_entries) if not bl])
    sum_entries = sum_true + sum_false
    if sum_entries == 0:
        # all nans scenario
        return np.nan
    # exclude nans from fraction
    return (1.0 * sum_true) / sum_entries


@Profiles.register(
    key="most_probable_value", description="Most probable value", dim=1, htype="all"
)
def profile_most_probable_value(bin_labels, bin_counts):
    return bin_labels[np.argmax(bin_counts)]



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from histogrammar.dfinterface.make_histograms import get_time_axes
from pydantic import BaseModel, BaseSettings
from typing_extensions import Literal

# Global configuration for the joblib parallelization. Could be used to change the number of jobs, and/or change
# the backend from default (loki) to 'multiprocessing' or 'threading'.
# (see https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html for details)
parallel_args = {"n_jobs": 1}


class ValidatedSettings(BaseSettings):
    class Config:
        validate_all = True
        validate_assignment = True


class SectionModel(BaseModel):
    name: str
    """Name of the section in the report"""

    description: str
    """Description of the section in the report"""


class ProfilesSection(SectionModel):
    name: str = "Profiles"
    """Name of the profiles section in the report"""

    description: str = """Basic statistics of the data (profiles) calculated for each time period (a period
                       is represented by one bin). The yellow and red lines represent the corresponding
                       traffic light bounds (default: 4 and 7 standard deviations with respect to the reference data)."""
    """Description of the profiles section in the report"""


class AlertSection(SectionModel):
    name: str = "Alerts"
    """Name of the alerts section in the report"""

    description: str = "Alerts aggregated by all traffic lights for each feature."
    """Description of the alerts section in the report"""

    descriptions: Dict[Literal["n_green", "n_yellow", "n_red"], str] = {
        "n_green": "Total number of green traffic lights (observed for all statistics)",
        "n_yellow": "Total number of  yellow traffic lights (observed for all statistics)",
        "n_red": "Total number of red traffic lights (observed for all statistics)",
    }
    """Descriptions of the individual alerts"""


class HistogramSectionModel(SectionModel):
    name: str = "Histograms"
    """Name of the histograms section in the report"""

    description: str = "This section contains visualisations of individual histograms and heatmaps of them over time."
    """Description of the histograms section in the report"""

    inspector_histogram_choices: int = 2
    """The number of histograms that can be compared at once (e.g. the number of dropdowns)"""

    hist_names: List[
        Literal["heatmap", "heatmap_column_normalized", "heatmap_row_normalized"]
    ] = [
        "heatmap",
        "heatmap_column_normalized",
        "heatmap_row_normalized",
    ]
    """Heatmaps of histograms to display in the report"""

    hist_names_formatted: Dict[
        Literal["heatmap", "heatmap_column_normalized", "heatmap_row_normalized"], str
    ] = {
        "heatmap": "Heatmap",
        "heatmap_column_normalized": "Column-Normalized Heatmap",
        "heatmap_row_normalized": "Row-Normalized Heatmap",
    }
    """Pretty-print names for the heatmaps"""

    descriptions: Dict[
        Literal["heatmap", "heatmap_column_normalized", "heatmap_row_normalized"], str
    ] = {
        "heatmap": "The heatmap shows the frequency of each value over time. If a variable has a high number of distinct values"
        "(i.e. has a high cardinality), then the most frequent values are displayed and the remaining are grouped as 'Others'. "
        "The maximum number of values to should is configurable (default: 20).",
        "heatmap_column_normalized": "The column-normalized heatmap allows for comparing of time bins when the counts in each bin vary.",
        "heatmap_row_normalized": "The row-normalized heatmaps allows for monitoring one value over time.",
    }
    """Descriptions of the heatmaps in the report"""

    plot_hist_n: int = 0
    """plot histograms for last 'n' periods. default is 0 to show all (optional)"""

    top_n: int = 20
    """plot heatmap for top 'n' categories. default is 20 (optional)"""

    cmap: str = "ylorrd"
    """colormap for histogram heatmaps"""


class TrafficLightsSection(SectionModel):
    name: str = "Traffic Lights"
    """Name of the traffic lights section in the report"""

    description: str = "Traffic light calculation for different statistics (based on the calculated normalized residual, a.k.a. pull). Statistics for which all traffic lights are green are hidden from view by default."
    """Description of the traffic lights section in the report"""


class ComparisonsSection(SectionModel):
    name: str = "Comparisons"
    """Name of the comparisons section in the report"""

    description: str = (
        "Statistical comparisons of each time period (one bin) to the reference data."
    )
    """Description of the comparisons section in the report"""


class OverviewSection(SectionModel):
    name: str = "Overview"
    """Name of the overview section in the report"""

    description: str = "Alerts aggregated per feature"
    """Description of the overview section in the report"""


class Section(BaseModel):
    """Configuration for the individual sections"""

    profiles: ProfilesSection = ProfilesSection()
    """Configuration related to the profiles section"""

    alerts: AlertSection = AlertSection()
    """Configuration related to the alerts section"""

    histograms: HistogramSectionModel = HistogramSectionModel()
    """Configuration related to the histogram section"""

    overview: OverviewSection = OverviewSection()
    """Configuration related to the overview section"""

    comparisons: ComparisonsSection = ComparisonsSection()
    """Configuration related to the comparisons section"""

    traffic_lights: TrafficLightsSection = TrafficLightsSection()
    """Configuration related to the traffic lights section"""


class Report(BaseModel):
    """Report-specific configuration"""

    title: str = "POPMON Report"
    """Report title in browser and navbar. May contain HTML."""

    last_n: int = 0
    """plot statistic data for last 'n' periods (optional)"""

    skip_first_n: int = 0
    """in plot skip first 'n' periods. last_n takes precedence (optional)"""

    skip_last_n: int = 0
    """in plot skip last 'n' periods. last_n takes precedence (optional)"""

    report_filepath: Optional[Union[str, Path]] = None
    """the file path where to output the report (optional)"""

    extended_report: bool = True
    """if True, show all the generated statistics in the report (optional)
    if set to False, then smaller show_stats (see below)"""

    online_report: bool = True
    """Use a CDN to host resources, or embed them into the report."""

    show_stats: List[str] = [
        "distinct*",
        "filled*",
        "nan*",
        "mean*",
        "std*",
        "p05*",
        "p50*",
        "p95*",
        "max*",
        "min*",
        "fraction_true*",
        "phik*",
        "*unknown_labels*",
        "*chi2_norm*",
        "*zscore*",
        "n_*",
        "*jsd*",
        "*psi*",
        "*max_prob_diff*",
    ]
    """list of statistic name patterns to show in the report. If None, show all (optional)"""

    primary_color: str = "#000080"
    """Primary color used throughout the report"""

    tl_colors: Dict[str, str] = {
        "green": "#008000",
        "yellow": "#FFC800",
        "red": "#FF0000",
    }
    """"Configure line colors in barplots of Comparisons and Profiles section. Need to be hex format (full length)"""

    section: Section = Section()
    """Configuration for the individual sections"""


class Comparison(BaseModel):
    """Parameters related to comparisons"""

    window: int = 10
    """size of rolling window and/or trend detection. default is 10."""

    shift: int = 1
    """shift of time-bins in rolling/expanding window. default is 1."""


class Monitoring(BaseModel):
    """Parameters related to monitoring"""

    monitoring_rules: Dict[str, List[Union[float, int]]] = {
        "*_pull": [7, 4, -4, -7],
        "*_zscore": [7, 4, -4, -7],
        "[!p]*_unknown_labels": [0.5, 0.5, 0, 0],
    }
    """
    monitoring rules to generate traffic light alerts.
    The default setting is:

    .. code-block:: python

        monitoring_rules = {
            "*_pull": [7, 4, -4, -7],
            "*_zscore": [7, 4, -4, -7],
            "[!p]*_unknown_labels": [0.5, 0.5, 0, 0],
        }

    Note that the (filename based) wildcards such as * apply to all statistic names matching that pattern.
    For example, ``"*_pull"`` applies for all features to all statistics ending on "_pull".
    You can also specify rules for specific features and/or statistics by leaving out wildcard and putting the
    feature name in front. E.g.

    .. code-block:: python

        monitoring_rules = {
            "featureA:*_pull": [5, 3, -3, -5],
            "featureA:nan": [4, 1, 0, 0],
            "*_pull": [7, 4, -4, -7],
            "nan": [8, 1, 0, 0],
        }

    In case of multiple rules could apply for a feature's statistic, the most specific one applies.
    So in case of the statistic "nan": "featureA:nan" is used for "featureA", and the other "nan" rule
    for all other features.
    """

    pull_rules: Dict[str, List[Union[float, int]]] = {"*_pull": [7, 4, -4, -7]}
    """
    red and yellow (possibly dynamic) boundaries shown in plots in the report.
    Default is:

    .. code-block:: python

        pull_rules = {"*_pull": [7, 4, -4, -7]}

    This means that the shown yellow boundaries are at -4, +4 standard deviations around the (reference) mean,
    and the shown red boundaries are at -7, +7 standard deviations around the (reference) mean.
    Note that the (filename based) wildcards such as * apply to all statistic names matching that pattern.
    (The same string logic applies as for monitoring_rules.)
    """


class Settings(ValidatedSettings):
    report: Report = Report()
    """Settings regarding the report"""

    comparison: Comparison = Comparison()
    """Settings related to the comparisons"""

    monitoring: Monitoring = Monitoring()
    """Settings related to monitoring"""

    time_axis: str = ""
    """
    name of datetime feature, used as time axis, e.g. 'date'. (column should be timestamp, date(time) or numeric batch id)
    if empty string, will be auto-guessed.
    """

    reference_type: Literal[
        "self", "external", "rolling", "expanding", "self_split"
    ] = "self"
    """
    type of reference used for comparisons
    """

    features: Optional[List[str]] = None
    """
    columns to pick up from input data. (default is all features).
    For multi-dimensional histograms, separate the column names with a ':'. Example features list is:

        .. code-block:: python

            features = ["x", "date", "date:x", "date:y", "date:x:y"]

    If time_axis is set or found, and if no features provided, features becomes: ['date:x', 'date:y', 'date:z'] etc.
    """

    binning: Literal["auto", "unit"] = "auto"
    """
    default binning to revert to in case bin_specs not supplied. When using "auto", semi-clever binning
    is automatically done.
    """

    bin_specs: Dict[str, Any] = {}
    """
    dictionaries used for rebinning numeric or timestamp features.
    An example bin_specs dictionary is:

    .. code-block:: python

        bin_specs = {
            "x": {"bin_width": 1, "bin_offset": 0},
            "y": {"num": 10, "low": 0.0, "high": 2.0},
            "x:y": [{}, {"num": 5, "low": 0.0, "high": 1.0}],
        }

    In the bin specs for x:y, x is not provided (here) and reverts to the 1-dim setting.
    The 'bin_width', 'bin_offset' notation makes an open-ended histogram (for that feature) with given bin width
    and offset. The notation 'num', 'low', 'high' gives a fixed range histogram from 'low' to 'high' with 'num'
    number of bins.
    """

    # Config utilities
    def _ensure_features_time_axis(self):
        self.features = [
            c if c.startswith(self.time_axis) else f"{self.time_axis}:{c}"
            for c in self.features
        ]

    def _set_time_axis_dataframe(self, df):
        time_axes = get_time_axes(df)
        num = len(time_axes)
        if num == 1:
            self.time_axis = time_axes[0]
        elif num == 0:
            raise ValueError(
                "No obvious time-axes found. Cannot generate stability report."
            )
        else:
            raise ValueError(
                f"Found {num} time-axes: {time_axes}. Set *one* time_axis manually!"
            )

    def _set_time_axis_hists(self, hists):
        # auto guess the time_axis: find the most frequent first column name in the histograms list
        first_cols = [k.split(":")[0] for k in list(hists.keys())]
        self.time_axis = max(set(first_cols), key=first_cols.count)

    def _set_bin_specs_by_time_width_and_offset(
        self, time_width: Union[str, int, float], time_offset: Union[str, int, float]
    ):
        if self.time_axis in self.bin_specs:
            raise ValueError(
                f'time-axis "{self.time_axis}" already found in binning specifications.'
            )
        # convert time width and offset to nanoseconds
        self.bin_specs[self.time_axis] = {
            "bin_width": float(pd.Timedelta(time_width).value),
            "bin_offset": float(pd.Timestamp(time_offset).value),
        }



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import logging

from popmon import Settings, resources
from popmon.base import Pipeline
from popmon.io import JsonReader
from popmon.pipeline.report_pipelines import SelfReference


class AmazingPipeline(Pipeline):
    def __init__(self, histogram_path: str, **kwargs):
        modules = [
            JsonReader(file_path=histogram_path, store_key=kwargs["hists_key"]),
            # Or ExternalReference, RollingReference etc.
            SelfReference(**kwargs),
        ]
        super().__init__(modules)


def run():
    """Example that run self-reference pipeline and produces monitoring report"""
    logging.basicConfig(
        level=logging.INFO, format="%(asctime)s %(levelname)s [%(module)s]: %(message)s"
    )

    settings = Settings(time_axis="date")
    settings.comparison.window = 20
    settings.comparison.shift = 1
    settings.monitoring.monitoring_rules = {
        "*_pull": [7, 4, -4, -7],
        # "*_pvalue": [1, 0.999, 0.001, 0.0001],
        "*_zscore": [7, 4, -4, -7],
    }
    settings.monitoring.pull_rules = {"*_pull": [7, 4, -4, -7]}

    histogram_path = resources.data("synthetic_histograms.json")
    pipeline = AmazingPipeline(histogram_path, hists_key="hists", settings=settings)
    pipeline.transform(datastore={})


if __name__ == "__main__":
    run()



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from typing import Optional

import pandas as pd
from tqdm import tqdm

from ..base import Module
from ..config import Report
from ..utils import filter_metrics, short_date
from .traffic_light_section_generator import _plot_metrics


class AlertSectionGenerator(Module):
    """This module takes the time-series data of already computed statistics, plots the data and
    combines all the plots into a list which is stored together with the section name in a dictionary
    which later will be used for the report generation.
    """

    _input_keys = ("read_key", "static_bounds", "dynamic_bounds", "store_key")
    _output_keys = ("store_key",)

    def __init__(
        self,
        read_key,
        store_key,
        settings: Report,
        features=None,
        ignore_features=None,
        static_bounds=None,
        dynamic_bounds=None,
        prefix="traffic_light_",
        suffices=["_red_high", "_yellow_high", "_yellow_low", "_red_low"],
        ignore_stat_endswith=None,
    ):
        """Initialize an instance of SectionGenerator.

        :param str read_key: key of input data to read from the datastore and use for plotting
        :param str store_key: key for output data to be stored in the datastore
        :param list features: list of features to pick up from input data (optional)
        :param list ignore_features: ignore list of features, if present (optional)
        :param str static_bounds: key to static traffic light bounds key in datastore (optional)
        :param str dynamic_bounds: key to dynamic traffic light bounds key in datastore (optional)
        :param str prefix: dynamic traffic light prefix. default is ``'traffic_light_'`` (optional)
        :param str suffices: dynamic traffic light suffices. (optional)
        :param list ignore_stat_endswith: ignore stats ending with any of list of suffices. (optional)
        """
        super().__init__()
        self.read_key = read_key
        self.store_key = store_key
        self.dynamic_bounds = dynamic_bounds
        self.static_bounds = static_bounds

        self.features = features or []
        self.ignore_features = ignore_features or []
        self.prefix = prefix
        self.suffices = suffices
        self.ignore_stat_endswith = ignore_stat_endswith or []

        self.last_n = settings.last_n
        self.skip_first_n = settings.skip_first_n
        self.skip_last_n = settings.skip_last_n
        self.show_stats = settings.show_stats if not settings.extended_report else None

        self.section_name = settings.section.alerts.name
        self.description = settings.section.alerts.description
        self.descriptions = settings.section.alerts.descriptions
        self.tl_colors = settings.tl_colors

    def get_description(self):
        return self.section_name

    def transform(
        self,
        data_obj: dict,
        static_bounds: Optional[dict] = None,
        dynamic_bounds: Optional[dict] = None,
        sections: Optional[list] = None,
    ):
        assert isinstance(data_obj, dict)
        if static_bounds is None:
            static_bounds = {}
        assert isinstance(static_bounds, dict)
        if dynamic_bounds is None:
            dynamic_bounds = {}
        assert isinstance(dynamic_bounds, dict)
        if sections is None:
            sections = []
        assert isinstance(sections, list)

        features = self.get_features(list(data_obj.keys()))
        features_w_metrics = []

        self.logger.info(f'Generating section "{self.section_name}"')

        for feature in tqdm(features, ncols=100):
            df = data_obj.get(feature, pd.DataFrame())
            fdbounds = dynamic_bounds.get(feature, pd.DataFrame(index=df.index))

            assert all(df.index == fdbounds.index)

            # prepare date labels
            df.drop(
                columns=["histogram", "reference_histogram"],
                inplace=True,
                errors="ignore",
            )
            dates = [short_date(date) for date in df.index.tolist()]

            metrics = filter_metrics(
                df.columns, self.ignore_stat_endswith, self.show_stats
            )

            plots = [
                _plot_metrics(
                    feature,
                    metrics,
                    dates,
                    df,
                    0,
                    0,
                    0,
                    self.tl_colors,
                    style="alerts",
                )
            ]

            # filter out potential empty plots (from skip empty plots)
            plots = [e for e in plots if len(e["plot"])]

            features_w_metrics.append(
                {
                    "name": feature,
                    "plot_type_layouts": {"traffic_lights": ""},
                    "plots": sorted(plots, key=lambda plot: plot["name"]),
                }
            )

        sections.append(
            {
                "section_title": self.section_name,
                "section_description": self.description,
                "features": features_w_metrics,
            }
        )
        return sections



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import logging
from abc import ABCMeta
from functools import wraps


def datastore_helper(func):
    """Decorator for passing and storing only the relevant keys in the datastore to
    the transform() method."""

    @wraps(func)
    def _transform(self, datastore):
        """Transformation helper function"""
        inputs = {}
        self.logger.debug(f"load from: {type(self)}")
        for key in self._input_keys:
            key_value = self.__dict__[key]
            if key_value and len(key_value) > 0:
                if isinstance(key_value, list):
                    inputs[key] = [datastore.get(k) for k in key_value]
                else:
                    inputs[key] = datastore.get(key_value)
            else:
                inputs[key] = None

            self.logger.debug(
                "load(key=%s, key_value=%s, value=%.100s})", key, key_value, inputs[key]
            )

        # transformation
        outputs = func(self, *list(inputs.values()))

        # transform returns None if no update needs to be made
        if outputs is not None:
            if len(self._output_keys) == 1:
                outputs = (outputs,)

            for k, v in zip(self._output_keys, outputs):
                key_value = self.__dict__[k]
                self.logger.debug(
                    "store(key=%s, key_value=%s, value=%.100s)", k, key_value, v
                )
                if key_value and len(key_value) > 0:
                    datastore[key_value] = v

        return datastore

    return _transform


class ModuleMetaClass(type):
    """Metaclass that wraps all transform() methods using the datastore_helper
    This obviates the need to decorate all methods in subclasses"""

    def __new__(cls, name, bases, local):
        if "transform" in local:
            value = local["transform"]
            if callable(value):
                local["transform"] = datastore_helper(value)
        return type.__new__(cls, name, bases, local)


def combine_classes(*args):
    """Combine multiple metaclasses"""
    name = "".join(a.__name__ for a in args)
    return type(name, args, {})


class Module(metaclass=combine_classes(ABCMeta, ModuleMetaClass)):
    """Abstract base class used for modules in a pipeline."""

    _input_keys = None
    _output_keys = None

    def __init__(self):
        """Module initialization"""
        self.logger = logging.getLogger()
        self.features = []
        self.feature_begins_with = []
        self.ignore_features = []

    def _get_values(self, keys):
        """Get the class attribute values for certain keys."""
        values = {}
        for x in keys:
            value = self.__dict__[x]
            if value != "" and value is not None and value not in values:
                values[x] = value
        return values

    def get_inputs(self):
        return self._get_values(self._input_keys)

    def get_outputs(self):
        return self._get_values(self._output_keys)

    def get_description(self):
        return ""

    def set_logger(self, logger):
        """Set logger of module

        :param logger: input logger
        """
        self.logger = logger

    @staticmethod
    def get_datastore_object(datastore, feature, dtype, default=None):
        """Get object from datastore.

        Bit more advanced than dict.get()

        :param dict datastore: input datastore
        :param str feature: key of object to retrieve
        :param obj dtype: required datatype of object. Could be specific data type or tuple of dtypes
        :param obj default: object to default to in case key not found.
        :return: retrieved object
        """
        if default is not None:
            obj = datastore.get(feature, default)
        else:
            try:
                obj = datastore[feature]
            except KeyError as e:
                raise ValueError(f"`{feature}` not found in the datastore!") from e

        if not isinstance(obj, dtype):
            raise TypeError(f"obj `{feature}` is not an instance of `{dtype}`!")
        return obj

    def get_features(self, all_features: list) -> list:
        """Get all features that meet feature_begins_with and ignore_features requirements

        :param list all_features: input features list
        :return: pruned features list
        :rtype: list
        """
        all_features = sorted(all_features)
        features = self.features or all_features

        if self.feature_begins_with:
            features = [k for k in features if k.startswith(self.feature_begins_with)]
        if self.ignore_features:
            features = [k for k in features if k not in self.ignore_features]

        features_not_in_input = [
            feature for feature in features if feature not in all_features
        ]
        for feature in features_not_in_input:
            self.logger.warning(f'Feature "{feature}" not in input data; skipping.')

        features = [feature for feature in features if feature in all_features]
        return features

    def transform(self, *args):
        """Central function of the module.

        Typically transform() takes something from the datastore, does something to it, and puts the results
        back into the datastore again, to be passed on to the next module in the pipeline.

        :param dict datastore: input datastore
        :return: updated output datastore
        :rtype: dict
        """
        raise NotImplementedError

    def __repr__(self):
        """String representation for modules when printing a pipeline/list of modules"""
        name = self.__class__.__name__
        input_keys = [f"{v}='{getattr(self, v)}'" for v in self._input_keys]
        output_keys = [f"{v}='{getattr(self, v)}'" for v in self._output_keys]
        params = ", ".join(input_keys + output_keys)
        return f"{name}({params})"



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import numpy as np
import pandas as pd


def mean(a, weights=None, axis=None, dtype=None, keepdims=False, ddof=0):
    """
    Compute the weighted mean along the specified axis.

    :param a: Array containing numbers whose mean is desired. If `a` is not an array, a conversion is attempted.
    :param weights: Array containing weights for the elements of `a`. If `weights` is not an
        array, a conversion is attempted.
    :param axis: Axis or axes along which the means are computed. The default is to
        compute the mean of the flattened array. Type is None or int or tuple of ints, optional.
    :param dtype: data type to use in computing the mean.
    :param bool keepdims: If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one.
    :param int ddof: delta degrees of freedom
    :return: np.ndarray
    """
    if weights is None:
        return np.mean(a, axis=axis, dtype=dtype, keepdims=keepdims)
    else:
        w = np.array(weights)

        return np.sum(w * np.array(a), axis=axis, dtype=dtype, keepdims=keepdims) / (
            np.sum(w, axis=axis, dtype=dtype, keepdims=keepdims) - ddof
        )


def std(a, weights=None, axis=None, dtype=None, ddof=0, keepdims=False):
    """
    Compute the weighted standard deviation along the specified axis.

    :param a: Array containing numbers whose standard deviation is desired. If `a` is not an
        array, a conversion is attempted.
    :param weights: Array containing weights for the elements of `a`. If `weights` is not an
        array, a conversion is attempted.
    :param axis: Axis or axes along which the means are computed. The default is to
        compute the mean of the flattened array. Type is None or int or tuple of ints, optional.
    :param dtype: data type to use in computing the mean.
    :param int ddof: Delta Degrees of Freedom.  The divisor used in calculations
        is ``W - ddof``, where ``W`` is the sum of weights (or number of elements
        if `weights` is None). By default `ddof` is zero
    :param bool keepdims: If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one.
    :return: np.ndarray
    """
    if weights is None:
        return np.std(a, axis=axis, dtype=dtype, ddof=ddof, keepdims=keepdims)
    else:
        m = mean(a, weights=weights, axis=axis, keepdims=True)
        v = mean((a - m) ** 2, weights=weights, axis=axis, keepdims=keepdims, ddof=ddof)
        return np.sqrt(v)


def median(a, weights=None, axis=None, keepdims=False):
    """
    Compute the weighted median along the specified axis.

    After https://en.wikipedia.org/wiki/Percentile#Weighted_percentile

    :param a: Array containing numbers whose median is desired. If `a` is not an
        array, a conversion is attempted.
    :param weights: Array containing weights for the elements of `a`. If `weights` is not an
        array, a conversion is attempted.
    :param axis: Axis or axes along which the means are computed. The default is to
        compute the mean of the flattened array. Type is None or int or tuple of ints, optional.
    :param bool keepdims: If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one.
    :return: number or array
    """
    return quantile(a, q=0.5, weights=weights, axis=axis, keepdims=keepdims)


def quantile(a, q, weights=None, axis=None, keepdims: bool = False):
    """
    Compute the weighted quantiles along the specified axis

    After https://en.wikipedia.org/wiki/Percentile#Weighted_percentile

    If `q` is a single quantile and `axis=None`, then the result
    is a scalar. If multiple quantiles are given, first axis of
    the result corresponds to the quantiles. The other axes are
    the axes that remain after the reduction of `a`.

    :param a: Array containing numbers whose median is desired. If `a` is not an
        array, a conversion is attempted
    :param q: Quantile or sequence of quantiles to compute, which must be between 0 and 1 inclusive
    :param weights: Array containing weights for the elements of `a`. If `weights` is not an
        array, a conversion is attempted.
    :param axis: Axis or axes along which the quantiles are computed. The
        default is to compute the quantile(s) along a flattened. Type is int, tuple of int, None, optional.
        version of the array
    :param bool keepdims: If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one.
    :return: scalar or ndarray
    """
    q = q if not hasattr(q, "__iter__") else q[0] if len(q) == 1 else tuple(q)
    if weights is None:
        return np.quantile(a, q, axis=axis, keepdims=keepdims)
    elif axis is None:
        raveled_data = np.ravel(a)
        idx = np.argsort(raveled_data)
        sorted_data = raveled_data[idx]
        sorted_weights = np.ravel(weights)[idx]
        Sn = np.cumsum(sorted_weights)
        Pn = (Sn - 0.5 * sorted_weights) / Sn[-1]
        y = np.interp(q, Pn, sorted_data)
        if keepdims:
            y = y.reshape((*y.shape, *(1,) * np.ndim(a)))

        return y
    else:
        # Move the dimensions which are reduced to the back
        axis = [axis] if not hasattr(axis, "__iter__") else axis
        destination = list(range(-len(axis), 0, 1))
        a_moved = np.moveaxis(a, source=axis, destination=destination)

        # Reshape into a 2D-array, with the first axis the dimensions
        # that are not reduced, and the second the dimensions that are reduced
        shape = (-1, np.prod(a_moved.shape[-len(axis) :]))
        a_shaped = a_moved.reshape(shape)

        w = np.moveaxis(weights, source=axis, destination=destination).reshape(shape)

        # Determine the quantiles and reshape backwards
        y = np.array([quantile(x, q, u) for x, u in zip(a_shaped, w)]).T
        if keepdims:
            shape = (
                *y.shape[:-1],
                *(1 if i in axis else x for i, x in enumerate(a.shape)),
            )
        else:
            shape = *y.shape[:-1], *a_moved.shape[: -len(destination)]

        y = y.reshape(shape)
        return y


def probability_distribution_mean_covariance(entries_list):
    """Mean normalized histogram and covariance of list of input histograms

    :param entries_list: numpy 2D array shape (n_histos, n_bins,) with bin counts of histograms
    :return: mean normalized histogram, covariance probability matrix
    """
    if len(entries_list) == 0:
        raise ValueError("List of input histogram entries is empty.")

    entries_list = np.atleast_2d(entries_list)
    n_histos = entries_list.shape[0]

    if n_histos == 1:
        # catch potential empty histogram
        if np.sum(entries_list[0]) == 0:
            return entries_list[0], None
        norm_hist_mean = entries_list[0] / np.sum(entries_list[0])
        return norm_hist_mean, None

    # At least two histograms from here on ...
    # Normalize the histograms along the bin axis, so that histograms with different number of entries
    # are still comparable
    normed_list = entries_list / (
        np.sum(entries_list, axis=1, dtype=float)[:, np.newaxis] + np.finfo(float).eps
    )

    # Determine the mean histogram (unbiased)
    norm_hist_mean = np.sum(normed_list, axis=0) / n_histos

    # For each histogram determine the second moment (i.e. mean of the product of two bins entries)
    # of all the other histograms
    cross_entries = normed_list[:, :, np.newaxis] * normed_list[:, np.newaxis, :]
    sum2_cross_entries = np.sum(cross_entries, axis=0) / n_histos

    # Determine the unbiased covariance matrices between bins for all the histograms.
    # note: use one degree of freedom less because of we're using the evaluated mean as input
    norm_hist_cov = (
        sum2_cross_entries
        - norm_hist_mean[:, np.newaxis] * norm_hist_mean[np.newaxis, :]
    ) * (n_histos / (n_histos - 1))

    return norm_hist_mean, norm_hist_cov


def covariance_multinomial_probability_distribution(entries):
    """Calculate covariance matrix of a single multinomial probability distribution

    :param entries: entries of input histogram
    :return: numpy 2D array with covariance matrix of multinomial probability distribution
    """
    n_bins = len(entries)
    n_entries = np.sum(entries)
    prob = entries / n_entries

    covariance_matrix = np.zeros((n_bins, n_bins))

    for i in range(n_bins):
        for j in range(i, n_bins):
            if i == j:
                covariance_matrix[i][j] = (prob[i] * (1 - prob[i])) / n_entries
            else:
                covariance_matrix[i][j] = -(prob[i] * prob[j]) / n_entries
                covariance_matrix[j][i] = covariance_matrix[i][j]

    return covariance_matrix


def mad(a, c=0.6745, axis=0):
    """Median Absolute Deviation along given axis of an array

    mad = median(abs(a - median(a)))/c

    Copyright statsmodels:
    Kindly taken from statsmodels package and then modified to work with dataframes as well.
    Reference: https://www.statsmodels.org/dev/_modules/statsmodels/robust/scale.html#mad
    License: https://github.com/statsmodels/statsmodels/blob/master/LICENSE.txt
    All modifications copyright ING WBAA.

    :param a: array_like Input array.
    :param float c: optional. The normalization constant. Defined as scipy.stats.norm.ppf(3/4.),
        which is approximately .6745.
    :param int axis: optional. The default is 0. Can also be None.
    :param center: callable or float. If a callable is provided, such as the default `np.median` then it
        is expected to be called center(a). The axis argument will be applied
        via np.apply_over_axes. Otherwise, provide a float.
    :return: mad
    :rtype: float
    """
    if isinstance(a, pd.DataFrame):
        a = a.select_dtypes([np.number]).dropna(axis=1, how="all")

    center = a.median(axis=axis)
    rel_abs_diff = (a - center).abs() / c
    mad = rel_abs_diff.median(axis=axis)
    return mad



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import warnings

import histogrammar
import numpy as np
from histogrammar.util import get_hist_props

from ..hist.hist_utils import get_bin_centers, is_numeric
from ..stats.numpy import quantile

used_hist_types = (histogrammar.Bin, histogrammar.SparselyBin, histogrammar.Categorize)


def prepare_ndgrid(hist, n_dim):
    """Get lists of all unique combinations of keys

    Used as input by get_ndgrid(hist).

    :param hist: input histogrammar histogram
    :return: n sorted lists of keys
    """

    if hist.n_dim < n_dim:
        warnings.warn(
            f"Input histogram only has {hist.n_dim} dimensions (<{n_dim}). Returning empty lists."
        )
        return [[] for _ in range(n_dim)]

    def get_hist_keys(h):
        if hasattr(h, "bins"):
            return set(h.bins.keys())
        elif hasattr(h, "values"):
            return set(range(len(h.values)))
        else:
            raise TypeError()

    # SparselyBin or Categorize
    def keys_recursive(hist, hist_keys, idx):
        hist_keys[idx] |= get_hist_keys(hist)
        if (idx + 1) < len(hist_keys):
            if hasattr(hist, "bins"):
                for h in hist.bins.values():
                    hist_keys = keys_recursive(h, hist_keys, idx + 1)
            elif hasattr(hist, "values"):
                for h in hist.values:
                    hist_keys = keys_recursive(h, hist_keys, idx + 1)
            else:
                raise TypeError()
        return hist_keys

    keys = [set() for _ in range(n_dim)]
    keys = keys_recursive(hist, keys, 0)
    keys = [sorted(k) for k in keys]
    return keys


def prepare_2dgrid(hist):
    """Get lists of all unique x and y keys

    Used as input by get_2dgrid(hist).

    :param hist: input histogrammar histogram
    :return: two sorted lists of unique x and y keys
    """
    return prepare_ndgrid(hist, n_dim=2)


def set_ndgrid(hist, keys, n_dim):
    """Set n-d grid of first n dimensions of input histogram

    Used as input by get_ndgrid(hist).

    :param hist: input histogrammar histogram
    :param list keys: list with unique keys per dim
    :return: filled nd numpy grid
    """
    shape = [len(k) for k in reversed(keys)]
    grid = np.zeros(shape)

    if hist.n_dim < n_dim:
        warnings.warn(
            f"Input histogram only has {hist.n_dim} dimensions (<{n_dim}). Returning empty grid."
        )
        return grid

    def flatten(histogram, keys, grid, dim=0, prefix=None):
        if prefix is None:
            prefix = []

        if len(keys) == len(prefix):
            grid[tuple(prefix)] = histogram.entries
        else:
            if hasattr(histogram, "bins"):
                for k, h in histogram.bins.items():
                    if k not in keys[dim]:
                        continue
                    i = keys[dim].index(k)
                    flatten(h, keys, grid, dim + 1, [i] + prefix)
            elif hasattr(histogram, "values"):
                for i, h in enumerate(histogram.values):
                    flatten(h, keys, grid, dim + 1, [i] + prefix)
            else:
                raise TypeError()

    flatten(hist, keys, grid)
    return grid


def set_2dgrid(hist, keys):
    """Set 2d grid of first two dimenstions of input histogram

    Used as input by get_2dgrid(hist).

    :param hist: input histogrammar histogram
    :param list xkeys: list with unique x keys
    :param list ykeys: list with unique y keys
    :return: filled 2d numpy grid
    """
    return set_ndgrid(hist, keys, n_dim=2)


def get_ndgrid(hist, get_bin_labels=False, n_dim=2):
    """Get filled n-d grid of first n dimensions of input histogram

    :param hist: input histogrammar histogram
    :return: grid of first n dimenstions of input histogram
    """
    if hist.n_dim < n_dim:
        warnings.warn(
            f"Input histogram only has {hist.n_dim} dimensions (<{n_dim}). Returning empty grid."
        )
        return np.zeros(tuple([0] * n_dim))

    keys = prepare_ndgrid(hist, n_dim)
    grid = set_ndgrid(hist, keys, n_dim)

    if get_bin_labels:
        return grid, keys

    return grid


def get_2dgrid(hist, get_bin_labels=False):
    """Get filled x,y grid of first two dimensions of input histogram

    :param hist: input histogrammar histogram
    :return: x,y grid of first two dimenstions of input histogram
    """
    return get_ndgrid(hist, get_bin_labels, n_dim=2)


def get_consistent_numpy_ndgrids(hist_list=[], get_bin_labels=False, dim=3):
    """Get list of consistent x,y grids of first n dimensions of (sparse) input histograms

    :param list hist_list: list of input histogrammar histograms
    :param bool get_bin_labels: if true, return x-keys and y-keys describing binnings of 2d-grid.
    :param int dim: number of dimension (>= 3)
    :return: list of consistent x,y grids of first two dimensions of each input histogram in list
    """
    # --- basic checks
    if len(hist_list) == 0:
        raise ValueError("Input histogram list has zero length.")
    if hist_list[0].n_dim < dim:
        raise ValueError(
            f"Input histogram only has {hist_list[0].n_dim} dimensions (<{dim}). Cannot compute {dim}d-grid."
        )
    assert_similar_hists(hist_list)

    keys = [set() for _ in range(dim)]
    for hist in hist_list:
        hist_keys = prepare_ndgrid(hist, n_dim=dim)
        for i, h_keys in enumerate(hist_keys):
            keys[i] |= set(h_keys)
    keys = [sorted(k) for k in keys]

    gridnd_list = [set_ndgrid(hist, keys, n_dim=dim) for hist in hist_list]

    if get_bin_labels:
        return gridnd_list, keys

    return gridnd_list


def get_consistent_numpy_2dgrids(hist_list=[], get_bin_labels=False):
    """Get list of consistent x,y grids of first two dimensions of (sparse) input histograms

    :param list hist_list: list of input histogrammar histograms
    :param bool get_bin_labels: if true, return x-keys and y-keys describing binnings of 2d-grid.
    :return: list of consistent x,y grids of first two dimensions of each input histogram in list
    """
    return get_consistent_numpy_ndgrids(hist_list, get_bin_labels, dim=2)


def get_consistent_numpy_1dhists(hist_list, get_bin_labels=False, crop_range=False):
    """Get list of consistent numpy hists for list of sparse (or bin) input histograms

    Works for sparse and bin histograms.
    Note: for sparse histograms, all potential bins between low and high are picked up (also unfilled).

    Note: a numpy histogram is a union of lists of bin_edges and number of entries
    This gives the full range of bin_centers, including zeros, which is not robust against (extreme) outliers.
    Ideally, use this for plotting of multiple histograms only.

    :param list hist_list: list of input histogram objects
    :param bool get_bin_labels: return bin labels as well, default is false.
    :param bool crop_range: return a trimmed version of the histogram, between 5-95% quantiles.
    :return: list of consistent 1d numpy hists for list of sparse input histograms
    """
    # --- basic checks
    if len(hist_list) == 0:
        raise ValueError("Input histogram list has zero length.")
    assert_similar_hists(hist_list)

    low_arr = [hist.low for hist in hist_list if hist.low is not None]
    high_arr = [hist.high for hist in hist_list if hist.high is not None]

    low = min(low_arr) if len(low_arr) > 0 else None
    high = max(high_arr) if len(high_arr) > 0 else None
    # low == None and/or high == None can only happen when all input hists are empty.

    if crop_range:
        # crop_range option crops a histogram to reasonable range, e.g. for plotting, giving nice plots.
        # in particular this protects against outliers that distort the view on the core part of the distribution
        # range is quantiles 5-95% + 5% on both sides
        q05_arr = []
        q95_arr = []
        for hist in hist_list:
            bin_centers, values = get_bin_centers(hist)
            bin_entries = np.array([v.entries for v in values])
            qs = quantile(bin_centers, [0.05, 0.95], bin_entries)
            q05_arr.append(qs[0])
            q95_arr.append(qs[1])
        q05 = min(q05_arr) if len(q05_arr) > 0 else np.nan
        q95 = max(q95_arr) if len(q95_arr) > 0 else np.nan
        delta = q95 - q05
        var_min = q05 - (0.06 / 0.9) * delta
        var_max = q95 + (0.06 / 0.9) * delta
        if 0.0 < var_min < 0.2 * delta:
            var_min = 0.0
        elif -0.2 * delta < var_max < 0.0:
            var_max = 0.0
        if not np.isnan(var_min) and low is not None and var_min > low:
            low = var_min
        if not np.isnan(var_max) and high is not None and var_max < high:
            high = var_max

    # if one of the input histograms is sparse and empty, copy the bin-edges and bin-centers
    # from a filled histogram, and use empty bin-entries array
    # MB 20220601: note this gives the full range of bin_centers, which is not robust against (extreme) outliers
    #              get_consistent_numpy_entries() ignores all empty bins.
    bin_edges = [0.0, 1.0]
    bin_centers = [0.5]
    null_entries = [0.0]
    if low is not None and high is not None:
        for hist in hist_list:
            if hist.low is not None and hist.high is not None:
                bin_edges = hist.bin_edges(low, high)
                bin_centers = hist.bin_centers(low, high)
                null_entries = np.zeros(len(bin_centers))
                break

    nphist_list = []
    for hist in hist_list:
        bin_entries = (
            null_entries
            if (hist.low is None and hist.high is None)
            else hist.bin_entries(low, high)
        )
        nphist_list.append((bin_entries, bin_edges))

    if get_bin_labels:
        return nphist_list, bin_centers
    else:
        return nphist_list


def get_consistent_numpy_entries(hist_list, get_bin_labels=False):
    """Get list of consistent numpy bin_entries for list of 1d input histograms

    Works for categorize, sparse and bin histograms.
    Note: for sparse histograms, *only* the filled bins are picked up.
    (this is not the case when calling get_consistent_numpy_1dhists(), which takes all bins b/n low and high.)

    :param list hist_list: list of input histogrammar histograms
    :return: list of consistent 1d numpy arrays with bin_entries for list of input histograms
    """
    # --- basic checks
    if len(hist_list) == 0:
        raise ValueError("Input histogram list has zero length.")
    assert_similar_hists(hist_list)

    # datatype check
    v0 = is_numeric(hist_list[0])
    constant = all(is_numeric(hist) == v0 for hist in hist_list[1:])
    all_num = constant and v0
    all_cat = constant and not v0
    if not (all_num or all_cat):
        raise TypeError(
            "Input histograms are mixture of Bin/SparselyBin and Categorize types."
        )

    # union of all labels encountered
    labels = set()
    for hist in hist_list:
        bin_labels = get_bin_centers(hist)[0]
        labels = labels.union(bin_labels)
    labels = sorted(labels)

    # collect list of consistent bin_entries
    if all_num:
        kwargs = {"xvalues": labels}
    else:
        # PATCH: deal with boolean labels, which get bin_labels() returns as strings
        cat_labels = labels
        props = get_hist_props(hist_list[0])
        if props["is_bool"]:
            cat_labels = [lab == "True" for lab in cat_labels]
        kwargs = {"labels": cat_labels}

    entries_list = [hist.bin_entries(**kwargs) for hist in hist_list]

    if get_bin_labels:
        return entries_list, labels
    else:
        return entries_list


def get_contentType(hist):
    """Get content type of bins of histogram

    :param hist: input histogram
    :return: string describing content type
    """
    if isinstance(hist, histogrammar.Count):
        return "Count"
    elif isinstance(hist, histogrammar.Bin):
        return "Bin"
    elif isinstance(hist, histogrammar.SparselyBin):
        return "SparselyBin"
    elif isinstance(hist, histogrammar.Categorize):
        return "Categorize"
    return "Count"


def check_similar_hists(hist_list, check_type=True, assert_type=used_hist_types):
    """Check consistent list of input histograms

    Check that type and dimension of all histograms in input list are the same.

    :param list hist_list: list of input histogram objects to check on consistency
    :param bool check_type: if true, also check type consistency of histograms (besides n-dim and datatype).
    :return: bool indicating if lists are similar
    """
    if len(hist_list) < 1:
        return True
    for hist in hist_list:
        if not isinstance(hist, assert_type):
            raise TypeError(
                "Input histogram type {htype} not of {htypes}.".format(
                    htype=type(hist), htypes=assert_type
                )
            )
    # perform similarity checks on:
    # - number of dimensions
    # - histogram type
    # - datatypes
    # - Bin attributes
    # - SparselyBin attributes
    # - all above for sub-histograms in case of n-dim > 1

    # Check generic attributes - filled histograms only
    n_d = [hist.n_dim for hist in hist_list]
    if not n_d.count(n_d[0]) == len(n_d):
        warnings.warn("Input histograms have inconsistent dimensions.")
        return False
    dts = [hist.datatype for hist in hist_list]
    if not dts.count(dts[0]) == len(dts):
        warnings.warn(f"Input histograms have inconsistent datatypes: {dts}")
        return False
    # Check generic attributes
    if check_type:
        # E.g. histogrammar.specialized.CategorizeHistogramMethods and
        # histogrammar.primitives.categorize.Categorize are both of type hg.Categorize
        # Make this consistent first.
        types = [get_contentType(hist) for hist in hist_list]
        if not types.count(types[0]) == len(types):
            warnings.warn(
                "Input histograms have inconsistent class types: {types}".format(
                    types=types
                )
            )
            return False

    # Check Bin attributes
    if isinstance(hist_list[0], histogrammar.Bin):
        nums = [hist.num for hist in hist_list]
        if not nums.count(nums[0]) == len(nums):
            warnings.warn(
                "Input Bin histograms have inconsistent num attributes: {types}".format(
                    types=nums
                )
            )
            return False
        lows = [hist.low for hist in hist_list]
        if not lows.count(lows[0]) == len(lows):
            warnings.warn(
                "Input Bin histograms have inconsistent low attributes: {types}".format(
                    types=lows
                )
            )
            return False
        highs = [hist.high for hist in hist_list]
        if not highs.count(highs[0]) == len(highs):
            warnings.warn(
                "Input histograms have inconsistent high attributes: {types}".format(
                    types=highs
                )
            )
            return False

    # Check SparselyBin attributes
    if isinstance(hist_list[0], histogrammar.SparselyBin):
        origins = [hist.origin for hist in hist_list]
        if not origins.count(origins[0]) == len(origins):
            warnings.warn(
                "Input SparselyBin histograms have inconsistent origin attributes: {types}".format(
                    types=origins
                )
            )
            return False
        bws = [hist.binWidth for hist in hist_list]
        if not bws.count(bws[0]) == len(bws):
            warnings.warn(
                "Input SparselyBin histograms have inconsistent binWidth attributes: {types}".format(
                    types=bws
                )
            )
            return False

    # Check sub-histogram attributes
    if n_d[0] > 1:
        sub_hist_list = []
        # Categorize and SparselyBin
        if hasattr(hist_list[0], "bins"):
            for hist in hist_list:
                kys = list(hist.bins.keys())
                if len(kys) > 0:
                    sub_hist_list.append(hist.bins[kys[0]])
        # Bin
        elif hasattr(hist_list[0], "values"):
            for hist in hist_list:
                if hist.num > 0:
                    sub_hist_list.append(hist.values[0])
        # iterate down
        if not check_similar_hists(sub_hist_list):
            return False

    return True


def assert_similar_hists(hist_list, check_type=True, assert_type=used_hist_types):
    """Assert consistent list of input histograms

    Assert that type and dimension of all histograms in input list are the same.

    :param list hist_list: list of input histogram objects to check on consistency
    :param bool assert_type: if true, also assert type consistency of histograms (besides n-dim and datatype).
    """
    similar = check_similar_hists(
        hist_list, check_type=check_type, assert_type=assert_type
    )
    if not similar:
        raise ValueError("Input histograms are not all similar.")


def check_same_hists(hist1, hist2):
    """Check if two hists are the same

    :param hist1: input histogram 1
    :param hist2: input histogram 2
    :return: boolean, true if two histograms are the same
    """
    same = check_similar_hists([hist1, hist2])
    same &= hist1.entries == hist2.entries
    same &= hist1.n_bins == hist2.n_bins
    same &= hist1.quantity.name == hist2.quantity.name
    return same



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from collections import defaultdict
from typing import Callable, Dict, List, Optional, Tuple, Union


class Registry:
    _properties = ("dim", "htype")

    def __init__(self):
        self._keys: List[str] = []
        self._descriptions: Dict[str, str] = {}
        self._properties_to_func = defaultdict(lambda: defaultdict(dict))
        self._func_name_to_properties = {}

    def register(
        self,
        key: Union[str, List[str], Tuple[str]],
        description: Union[str, List[str], Tuple[str]],
        dim: int = -1,
        htype: Optional[str] = None,
    ):
        # rename for function use, without changing api
        keys = key
        del key

        descriptions = description
        del description

        # ensure that keys are a tuple
        if isinstance(keys, list):
            keys = tuple(keys)
        elif not isinstance(keys, tuple):
            keys = (keys,)

        # ensure that description is a tuple
        if isinstance(descriptions, list):
            descriptions = tuple(descriptions)
        elif not isinstance(descriptions, tuple):
            descriptions = (descriptions,)

        def f(func: Callable):
            # function names should be unique
            if func.__name__ in self._func_name_to_properties:
                raise ValueError(
                    f"A function with the name '{func.__name__}' has already been registered."
                )

            # keys should unique correspond to a function
            for key in keys:
                if key in self._keys:
                    raise ValueError(f"Key '{key}' has already been registered.")

            # register properties
            self._keys += list(keys)
            self._func_name_to_properties[func.__name__] = (dim, htype, keys)
            self._properties_to_func[dim][htype][keys] = func
            self._descriptions.update(dict(zip(keys, descriptions)))

            return func

        return f

    # Methods
    def _get_func_properties_by_name(
        self, function_name: str
    ) -> Tuple[int, str, Tuple[str]]:
        return self._func_name_to_properties[function_name]

    def get_func_by_name(self, function_name: str) -> Callable:
        """
         Get a function by the function name

        :param str function_name: name of the original function
        """
        dim, htype, key = self._get_func_properties_by_name(function_name)
        return self._properties_to_func[dim][htype][key]

    def get_func_by_dim_and_htype(self, dim, htype) -> Dict[Tuple[str], Callable]:
        return self._properties_to_func[dim][htype]

    def get_keys(self) -> List[str]:
        """List of keys associated with registered functions"""
        return self._keys

    def get_keys_by_dim_and_htype(self, dim, htype) -> List[str]:
        """Flat list of keys for a provided dimension and histogram type"""
        return [
            v for values in self._properties_to_func[dim][htype].keys() for v in values
        ]

    def get_descriptions(self) -> Dict[str, str]:
        """Dictionary of key->description associated with registered functions"""
        return self._descriptions

    def update_func(self, name, func) -> None:
        dim, htype, key = self._func_name_to_properties[name]
        self._properties_to_func[dim][htype][key] = func

    def run(self, args, dim, htype):
        output = {}
        for key, func in self.get_func_by_dim_and_htype(dim=dim, htype=htype).items():
            results = func(*args)
            if not isinstance(results, tuple):
                results = (results,)
            output.update(dict(zip(key, results)))
        return output



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import numpy as np
from scipy import stats

from popmon.base.registry import Registry

Comparisons = Registry()


@Comparisons.register(
    key="max_prob_diff",
    description="The largest absolute difference between all bin pairs of two normalized histograms",
    htype="all",
)
def googl_test(bins_1, bins_2):
    """Google-paper test

    Reference link: https://mlsys.org/Conferences/2019/doc/2019/167.pdf

    :param bins_1: first array of bin entries
    :param bins_2: second array of entries

    :return: maximum difference between the two entry distributions
    :rtype: float
    """

    def dist(bins):
        sum_ = np.sum(bins)
        return bins / sum_ if sum_ else bins

    return np.max(np.abs(dist(bins_1) - dist(bins_2)))


@Comparisons.register(key="psi", description="Population Stability Index", htype="all")
def population_stability_index(po, qo):
    epsilon = 10e-6
    p = po.copy()
    q = qo.copy()
    p += epsilon
    q += epsilon
    return np.sum((p - q) * np.log(p / q))


def kullback_leibler_divergence(po, qo):
    epsilon = 10e-6
    p = po.copy()
    q = qo.copy()
    p += epsilon
    q += epsilon
    return np.sum(p * np.log(p / q))


@Comparisons.register(key="jsd", description="Jensen-Shannon Divergence", htype="all")
def jensen_shannon_divergence(p, q):
    m = 0.5 * (p + q)
    return 0.5 * (kullback_leibler_divergence(p, m) + kullback_leibler_divergence(q, m))


def ks_test(hist_1, hist_2):
    """KS-test for two histograms with different number of entries

    Copyright ROOT:
    Formulas translated from c++ to python, but formulas otherwise not modified.
    Reference: link: https://root.cern.ch/doc/master/classTH1.html#TH1:KolmogorovTest
    GNU license: https://root.cern.ch/license
    All modifications copyright ING WBAA.

    :param hist_1: 1D array with bin counts of the histogram_1
    :param hist_2: 1D array with bin counts of the histogram_2

    :return: ks_score: Kolmogorov-Smirnov Test score
    :rtype: float
    """
    if len(hist_1) == 0 or len(hist_2) == 0:
        raise ValueError("Input histogram(s) has zero size.")
    if len(hist_1) != len(hist_2):
        raise ValueError("Input histograms have unequal size.")

    sum_1 = np.sum(hist_1)
    sum_2 = np.sum(hist_2)
    if sum_1 == 0 or sum_2 == 0:
        return np.nan

    normalized_cumsum_1 = np.cumsum(hist_1) / sum_1
    normalized_cumsum_2 = np.cumsum(hist_2) / sum_2

    d = np.abs(normalized_cumsum_1 - normalized_cumsum_2)

    return np.max(d) * np.sqrt(sum_1 * sum_2 / (sum_1 + sum_2))


def ks_prob(testscore):
    """KS-probability corresponding ti KS test score

    Copyright ROOT:
    Formulas translated from c++ to python, but formulas otherwise not modified.
    Reference: https://root.cern.ch/doc/master/classTH1.html#TH1:KolmogorovTest
    GNU license: https://root.cern.ch/license
    All modifications copyright ING WBAA.

    :param float testscore: Kolmogorov-Smirnov test score

    :return: approximate pvalue for the Kolmogorov-Smirnov test score
    :rtype: float
    """
    fj = np.array([-2, -8, -18, -32])
    r = np.zeros(4)

    w = 2.50662827
    c = np.array([-1.2337005501361697, -11.103304951225528, -30.842513753404244])

    u = abs(testscore)
    pvalue = np.nan
    if u < 0.2:
        pvalue = 1
    elif u < 0.755:
        v = np.power(u, -2)
        pvalue = 1 - w * np.exp(c * v).sum() / u
    elif u < 6.8116:
        v = np.power(u, 2)
        max_j = int(max(1, round(3.0 / u)))
        r[:max_j] = np.exp(fj[:max_j] * v)
        pvalue = 2 * (r[0] - r[1] + r[2] - r[3])

    return pvalue


@Comparisons.register(
    key=["ks", "ks_pvalue", "ks_zscore"],
    description=[
        "Kolmogorov-Smirnov test statistic",
        "p-value of the Kolmogorov-Smirnov test",
        "Z-score of the Kolmogorov-Smirnov test",
    ],
    dim=1,
    htype="num",
)
def ks(p, q, *args):
    # KS-test only properly defined for (ordered) 1D interval variables
    ks_testscore = ks_test(p, q)
    ks_pvalue = ks_prob(ks_testscore)
    ks_zscore = -stats.norm.ppf(ks_pvalue)
    return ks_testscore, ks_pvalue, ks_zscore


@Comparisons.register(
    key="unknown_labels",
    description="Are categories observed in a given time slot that are not present in the reference?",
    dim=1,
    htype="cat",
)
def unknown_labels(hist1, hist2):
    # check consistency of bin_labels
    labels1 = hist1.keySet
    labels2 = hist2.keySet
    subset = labels1 <= labels2
    return not subset


@Comparisons.register(
    key="pearson",
    description="Pearson correlation coefficient",
    dim=(2,),
    htype="all",
)
def pearson(p, q, *args):
    # calculate pearson coefficient
    pearson_coeff = np.nan
    if len(p) >= 2:
        same0 = all(p == p[0])
        same1 = all(q == q[0])
        if not same0 and not same1:
            # this avoids std==0, and thereby avoid runtime warnings
            pearson_coeff, _ = stats.pearsonr(p, q)
    return pearson_coeff


def uu_chi2(n, m):
    """Normalized Chi^2 formula for two histograms with different number of entries

    Copyright ROOT:
    Formulas translated from c++ to python, but formulas otherwise not modified.
    Reference: https://root.cern.ch/doc/master/classTH1.html#a6c281eebc0c0a848e7a0d620425090a5
    GNU License: https://root.cern.ch/license
    All modifications copyright ING WBAA.

    :param n: 1d array with bin counts of the reference set
    :param m: 1d array with bin counts of the test set
    :return: tuple of floats (chi2_value, chi2_norm, z_score, p_value, res)
    """

    def _not_finite_to_zero(x):
        res = x.copy()
        res[~np.isfinite(res)] = 0
        return res

    if len(n) == 0 or len(m) == 0:
        raise ValueError("Input histogram(s) has zero size.")
    if len(n) != len(m):
        raise ValueError("Input histograms have unequal size.")

    N = np.sum(n)
    M = np.sum(m)

    if N == 0 or M == 0:
        return np.nan, np.nan, np.nan, np.nan, [0] * len(n)

    # remove all zero entries in the sum, to present division by zero for individual bins
    z = n + m
    n = n[z != 0]
    m = m[z != 0]

    dof = ((n != 0) | (m != 0)).sum() - 1
    chi2_value = _not_finite_to_zero(((M * n - N * m) ** 2) / (n + m)).sum() / M / N

    chi2_norm = chi2_value / dof if dof > 0 else np.nan
    p_value = stats.chi2.sf(chi2_value, dof)
    z_score = -stats.norm.ppf(p_value)

    p = (n + m) / (N + M)

    if (p == 1).any():
        # unusual case of (only) one bin with p==1, avoids division with zero below
        res = np.array([np.nan] * len(p))
    else:
        res = _not_finite_to_zero(
            (n - N * p) / np.sqrt(N * p) / np.sqrt((1 - N / (N + M)) * (1 - p))
        )

    return chi2_value, chi2_norm, z_score, p_value, res


@Comparisons.register(
    key=[
        "chi2",
        "chi2_norm",
        "chi2_zscore",
        "chi2_pvalue",
        "chi2_max_residual",
        "chi2_spike_count",
    ],
    description=[
        "Chi-squared test statistic",
        "Normalized chi-squared statistic",
        "Z-score of the chi-squared statistic",
        "p-value of the chi-squared statistic",
        "The largest absolute normalized residual (|chi|) observed in all bin pairs",
        "The number of normalized residuals of all bin pairs with absolute value bigger than a given threshold (default: 7).",
    ],
    htype="all",
)
def chi2(*args, max_res_bound=7.0):
    chi2r, chi2_norm, zscore, pvalue, res = uu_chi2(*args)
    abs_residual = np.abs(res)
    chi2_max_residual = np.max(abs_residual)
    chi2_spike_count = np.sum(abs_residual[abs_residual > max_res_bound])

    return chi2r, chi2_norm, zscore, pvalue, chi2_max_residual, chi2_spike_count



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import json
import logging
import math
import warnings
from collections import defaultdict
from typing import Dict, List

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go

from popmon.resources import templates_env

NUM_NS_DAY = 24 * 3600 * int(1e9)

logger = logging.getLogger()


# set x-axis tick length
def xtick(lab, top):
    """Get x-tick."""
    lab = str(lab)
    if len(lab) > top:
        lab = lab[: top - 3] + "..."
    return lab


def plot_bars(
    data,
    labels: List[str],
    bounds: tuple,
    ylim: bool,
    primary_color: str,
    tl_colors: Dict[str, str],
    metric: str,
) -> str:
    """Plotting histogram data.

    :param numpy.ndarray data: bin values of a histogram
    :param labels: common bin labels for all histograms. default is None.
    :param bounds: traffic light bounds (y-coordinates). default is None.
    :param ylim: place y-axis limits for zooming into the data. default is False.
    :return: JSON plot image
    :rtype: str
    """
    # basic checks first
    n = data.size  # number of bins
    if labels is not None and len(labels) != n:
        raise ValueError("shape mismatch: x-axis labels do not match the data shape")

    # skip plot generation for empty datasets
    n_data = len(data)
    n_nan = pd.isnull(data).sum()
    n_inf = np.sum([np.isinf(x) for x in data if isinstance(x, float)])
    if n_nan + n_inf == n_data:
        logger.debug("skipping plot with empty data.")
        return ""

    # plot bar
    fig = go.Figure(
        [
            go.Bar(
                x=labels,
                y=data,
                hovertemplate="%{y:.4f}",
                name=metric,
                marker_color=primary_color,
            )
        ]
    )

    # set label granularity
    if len(labels) > 0:
        granularity = math.ceil(len(labels) / 50)
        labels = labels[::granularity]

    fig.update_layout(
        xaxis_tickangle=-90,
        xaxis={"type": "category"},
        margin={"l": 40, "r": 10, "t": 30},
    )
    fig.update_xaxes(
        tickvals=labels,
        ticktext=labels,
        showgrid=True,
        ticks="outside",
        minor_ticks="outside",
        showline=True,
        linecolor="black",
        mirror=True,
    )
    fig.update_yaxes(
        ticks="outside",
        minor_ticks="outside",
        showline=True,
        linecolor="black",
        mirror=True,
    )
    fig.update_traces(
        marker_line_color=primary_color,
        marker_line_width=1,
    )
    # plot boundaries
    try:
        all_nan = (np.isnan(data)).all()
        max_value = np.nanmax(data) if not all_nan else np.nan
        min_value = np.nanmin(data) if not all_nan else np.nan

        if len(bounds) > 0:
            max_r, max_y, min_y, min_r = bounds
            y_max = max(
                max(max_r) if isinstance(max_r, (list, tuple)) else max_r, max_value
            )
            y_min = min(
                max(min_r) if isinstance(min_r, (list, tuple)) else min_r, min_value
            )
            spread = (y_max - y_min) / 20
            y_max += spread
            y_min -= spread
            if not isinstance(max_r, (list, tuple)):
                fig.add_hline(y=max_r, line_color=tl_colors["red"])
                fig.add_hline(y=max_y, line_color=tl_colors["yellow"])
                fig.add_hline(y=min_y, line_color=tl_colors["yellow"])
                fig.add_hline(y=min_r, line_color=tl_colors["red"])
            else:
                fig.add_hline(y=max_r[0], line_color=tl_colors["red"])
                fig.add_hline(y=max_y[0], line_color=tl_colors["yellow"])
                fig.add_hline(y=min_y[0], line_color=tl_colors["yellow"])
                fig.add_hline(y=min_r[0], line_color=tl_colors["red"])

            if y_max > y_min:
                fig.update_yaxes(range=[y_min, y_max])

        elif ylim:
            spread = (max_value - min_value) / 20
            y_min = min_value - spread
            y_max = max_value + spread
            if y_max > y_min:
                fig.update_yaxes(range=[y_min, y_max])
    except Exception:
        logger.debug("unable to plot boundaries")

    plot = json.loads(fig.to_json())
    return plot


def plot_traffic_lights_overview(feature, data, metrics: List[str], labels: List[str]):
    colors = defaultdict(dict)
    color_map = ["g", "y", "r"]
    for c1, metric in enumerate(metrics):
        for c2, label in enumerate(labels):
            colors[metric][label] = [color_map[data[c1][c2]]]

    return templates_env(
        "table.html",
        feature=feature,
        metrics=metrics,
        labels=labels,
        data=colors,
        links=True,
    )


def hex_to_rgb(h):
    """Takes a hex rgb string and returns an RGB tuple."""
    return tuple(int(h[i : i + 2], 16) for i in (1, 3, 5))


def plot_traffic_lights_alerts_aggregate(
    feature, data, metrics: List[str], labels: List[str], tl_colors: Dict[str, str]
):
    assert data.shape[0] == 3

    # Reorder metrics if needed
    pos_green = metrics.index("n_green")
    pos_yellow = metrics.index("n_yellow")
    pos_red = metrics.index("n_red")

    if [pos_green, pos_yellow, pos_red] != [0, 1, 2]:
        data[[0, 1, 2]] = data[[pos_green, pos_yellow, pos_red]]

    metrics = ["# green", "# yellow", "# red"]
    data = data.astype(int)

    green = hex_to_rgb(tl_colors["green"])
    yellow = hex_to_rgb(tl_colors["yellow"])
    red = hex_to_rgb(tl_colors["red"])

    colors = defaultdict(dict)
    for c1, metric in enumerate(metrics):
        row_max = np.max(data[c1])
        for c2, label in enumerate(labels):
            a = np.round(data[c1][c2] / row_max, 2) if row_max and row_max != 0 else 0
            if metric.endswith("green"):
                background_rgba = (*green, a)
            elif metric.endswith("yellow"):
                background_rgba = (*yellow, a)
            else:
                background_rgba = (*red, a)
            background_rgba = (str(v) for v in background_rgba)
            text_color = "white" if a > 0.5 else "black"
            colors[metric][label] = (text_color, background_rgba, data[c1][c2])

    return templates_env(
        "table.html",
        feature=feature,
        metrics=metrics,
        labels=labels,
        data=colors,
        links=False,
    )


# basic checks for histograms
def histogram_basic_checks(plots={}):
    if len(plots) == 0:
        return

    for plot in plots:
        if len(plot["hist_names"]) == 0:
            plot["hist_names"] = [f"hist{i}" for i in range(len(plot["hists"]))]
        if plot["hist_names"]:
            if len(plot["hists"]) != len(plot["hist_names"]):
                raise ValueError("length of hist and hist_names are different")

        for i, hist in enumerate(plot["hists"]):
            try:
                hist_values, hist_bins = hist
            except BaseException as e:
                raise ValueError(
                    "Cannot extract binning and values from input histogram"
                ) from e

            assert hist_values is not None and len(
                hist_values
            ), "Histogram bin values have not been set."
            assert hist_bins is not None and len(
                hist_bins
            ), "Histogram binning has not been set."

            if plot["is_ts"]:
                plot["is_num"] = True

            if plot["is_num"]:
                bin_edges = hist_bins
                bin_values = hist_values
                assert (
                    len(bin_edges) == len(bin_values) + 1
                ), "bin edges (+ upper edge) and bin values have inconsistent lengths: {:d} vs {:d}. {}".format(
                    len(bin_edges), len(bin_values), plot["feature"]
                )
            else:
                labels = hist_bins
                values = hist_values
                assert len(labels) == len(
                    values
                ), f'labels and values have different array lengths: {len(labels):d} vs {len(values):d}. {plot["feature"]}'


def plot_histogram_overlay(
    plots=[],
    is_num=True,
    is_ts=False,
    is_static_reference=True,
    top=20,
    n_choices=2,
):
    """Create and plot (overlapping/grouped) histogram(s) of column values.

    Copyright Eskapade:
    Kindly taken from Eskapade package and then modified. Reference link:
    https://github.com/KaveIO/Eskapade/blob/master/python/eskapade/visualization/vis_utils.py#L397
    License: https://github.com/KaveIO/Eskapade-Core/blob/master/LICENSE
    Modifications copyright ING WBAA.

    :param list plots: list of dicts containing histograms for all timestamps
        :param bool is_num: True if observable to plot is numeric. default is True.
        :param bool is_ts: True if observable to plot is a timestamp. default is False.
    :param bool is_static_reference: True if the reference is static. default is True
    :param int top: only print the top 20 characters of x-labels and y-labels. default is 20.
    :param int n_choices: number of plots to compare at once
    :return: JSON encoded plot image
    :rtype: str
    """

    fig = go.Figure()

    alpha = 0.4

    # check number of plots
    if len(plots) < 2:
        warnings.warn("insufficient plots for histogram inspection")
        return

    base_plot = plots[0]

    # basic attribute check: time stamps treated as numeric.
    if is_ts:
        is_num = True

    # plot numeric and time stamps
    if is_num:

        # plot histogram
        for index in range(n_choices):
            bin_edges = plots[index]["hists"][0][1]
            bin_values = plots[index]["hists"][0][0]
            fig.add_trace(
                go.Bar(
                    x=bin_edges[1:],
                    y=bin_values,
                    opacity=alpha,
                    showlegend=True,
                    name=plots[index]["date"],
                    meta=index,
                )
            )

        # plot reference
        for index in range(1 if is_static_reference else n_choices):
            bin_edges = (
                plots[index]["hists"][0][1]
                if len(plots[index]["hists"]) < 2
                else plots[index]["hists"][1][1]
            )
            bin_values = (
                [0 for x in range(len(plots[index]["hists"][0][0]))]
                if len(plots[index]["hists"]) < 2
                else plots[index]["hists"][1][0]
            )
            fig.add_trace(
                go.Bar(
                    x=bin_edges[1:],
                    y=bin_values,
                    opacity=alpha,
                    showlegend=True,
                    name="no_ref"
                    if len(plots[index]["hists"]) < 2
                    else "Reference"
                    if is_static_reference
                    else (plots[index]["date"] + "-")
                    + plots[index]["hist_names"][1].split("_")[-1],
                    meta=index + 2,
                )
            )

        # set x-axis properties
        xlim = [min(bin_edges), max(bin_edges)]
        fig.update_xaxes(range=xlim)

    # plot categories
    else:

        # plot histogram for first 'n_choices' timestamps
        for index in range(n_choices):
            labels = plots[index]["hists"][0][1]
            values = plots[index]["hists"][0][0]
            fig.add_trace(
                go.Bar(
                    x=[xtick(lab, top) for lab in labels],
                    y=values,
                    opacity=alpha,
                    showlegend=True,
                    name=plots[index]["date"],
                    meta=index,
                )
            )

        # plot reference for first 1 or 'n_choices' timestamps
        for index in range(1 if is_static_reference else n_choices):
            labels = (
                plots[index]["hists"][0][1]
                if len(plots[index]["hists"]) < 2
                else plots[index]["hists"][1][1]
            )
            values = (
                [0 for _ in range(len(plots[index]["hists"][0][0]))]
                if len(plots[index]["hists"]) < 2
                else plots[index]["hists"][1][0]
            )
            fig.add_trace(
                go.Bar(
                    x=[xtick(lab, top) for lab in labels],
                    y=values,
                    opacity=alpha,
                    showlegend=True,
                    name="no_ref"
                    if len(plots[index]["hists"]) < 2
                    else "Reference"
                    if is_static_reference
                    else plots[index]["date"]
                    + " "
                    + plots[index]["hist_names"][1].split("_")[-1],
                    meta=index + n_choices,
                )
            )

    # set common histogram layout properties
    y_label = (
        str(base_plot["y_label"]) if base_plot["y_label"] is not None else "Bin count"
    )
    fig.update_yaxes(
        title=y_label,
        minor_ticks="outside",
        showline=True,
        linecolor="black",
        mirror=True,
    )
    fig.update_xaxes(
        title=base_plot["feature"],
        minor_ticks="outside",
        showline=True,
        linecolor="black",
        mirror=True,
    )
    fig.update_layout(
        barmode="overlay",
        legend={
            "orientation": "h",
            "yanchor": "bottom",
            "y": 1.1,
            "xanchor": "left",
            "x": 0,
            "font": {"size": 10},
        },
        hovermode="x unified",
        margin={"l": 40, "r": 10},
    )

    # dropdown menu
    fig.update_layout(
        updatemenus=[
            *[
                {
                    "buttons": [
                        {
                            "label": f'{plot["date"]}',
                            "method": "restyle",
                            "args": [
                                {
                                    "y": [
                                        plot["hists"][0][0],
                                        [0 for _ in range(len(plot["hists"][0][0]))]
                                        if len(plot["hists"]) < 2
                                        else plot["hists"][1][0],
                                    ],
                                    "name": [
                                        plot["date"],
                                        "no_ref"
                                        if len(plot["hist_names"]) < 2
                                        else "Reference"
                                        if is_static_reference
                                        else plots[index]["date"]
                                        + " "
                                        + plot["hist_names"][1].split("_")[-1],
                                    ],
                                },
                                [b, b + 2],
                            ],
                        }
                        for plot in plots
                    ],
                    "active": b,
                    "pad": {"r": 10, "t": 10},
                    "borderwidth": 0,
                    "bgcolor": "#d3d3d3",
                    "showactive": True,
                    "x": b / 5,
                    "y": 1.45,
                    "xanchor": "left",
                    "yanchor": "top",
                }
                for b in range(n_choices)
            ],
            {
                "buttons": [
                    {
                        "label": mode,
                        "method": "relayout",
                        "args": [
                            {
                                "barmode": mode,
                            }
                        ],
                    }
                    for mode in ["overlay", "group"]
                ],
                "pad": {"r": 10, "t": 10},
                "borderwidth": 0,
                "bgcolor": "#d3d3d3",
                "showactive": True,
                "x": 1,
                "y": 1.45,
                "xanchor": "right",
                "yanchor": "top",
            },
        ]
    )

    plot = json.loads(fig.to_json())
    return {
        "name": "Histogram Inspector ",
        "type": "histogram",
        "description": "",
        "plot": plot.get("data", ""),
        "layout": plot.get("layout", ""),
        "full_width": True,
    }


def plot_heatmap(
    hist_values: list,
    hist_bins: list,
    date: list,
    x_label: str,
    hist_name,
    y_label: str,
    is_num: bool = False,
    is_ts: bool = False,
    cmap: str = "ylorrd",
    top: int = 20,
):
    """Create and plot heatmap of column values.

    Copyright Eskapade:
    Kindly taken from Eskapade package and then modified. Reference link:
    https://github.com/KaveIO/Eskapade/blob/master/python/eskapade/visualization/vis_utils.py#L397
    License: https://github.com/KaveIO/Eskapade-Core/blob/master/LICENSE
    Modifications copyright ING WBAA.

    :param list hist_values: values of heatmap in a 2d numpy array =
    :param list hist_bins: bin labels/edges on y-axis
    :param list date: dates for x/time axis of heatmap
    :param str x_label: Label for heatmap x-axis
    :param list hist_names: list of histogram names. default is [].
    :param str y_label: Label for histogram y-axis. default is None.
    :param bool is_num: True if observable to plot is numeric. default is True.
    :param bool is_ts: True if observable to plot is a timestamp. default is False.
    :param int top: only print the top 20 characters of x-labels and y-labels. default is 20.
    :param cmap: the colormap for heatmap. default is ylorrd.
    :return: base64 encoded plot image
    :rtype: str
    """
    if hist_name:
        if len(hist_name) == 0:
            raise ValueError("length of heatmap names is zero")

    assert hist_values is not None and len(
        hist_values
    ), "Heatmap bin values have not been set."
    assert hist_bins is not None and len(hist_bins), "Heatmap binning has not been set."

    # basic attribute check: time stamps treated as numeric.
    if is_ts:
        is_num = True

    # plot numeric and time stamps
    if is_num:
        bin_edges = hist_bins
        bin_values = hist_values
        assert (
            len(bin_edges) == len(bin_values) + 1
        ), "bin edges (+ upper edge) and bin values have inconsistent lengths: {:d} vs {:d}. {}".format(
            len(bin_edges), len(bin_values), x_label
        )
        return ""

    # plot categories
    else:
        labels = hist_bins
        values = hist_values
        assert len(labels) == len(
            values
        ), "labels and values have different array lengths: {:d} vs {:d}. {}".format(
            len(labels), len(values), x_label
        )

        # plot histogram
        fig = px.imshow(
            values,
            labels={"x": "Time Bins", "y": x_label, "color": y_label},
            x=date,
            y=[xtick(lab, top) for lab in labels],
            color_continuous_scale=cmap,
            text_auto=".2f",
            aspect="equal",
        )

        # set label granularity
        if len(date) > 0:
            granularity = math.ceil(len(date) / 50)
            date = date[::granularity]

        fig.update_xaxes(tickvals=date, ticktext=date, tickangle=-90)
        fig.update_yaxes(ticks="outside")
        fig.update_layout(xaxis={"type": "category"}, margin={"l": 40, "r": 10, "t": 0})
        fig.update_coloraxes(colorbar_len=0.8, colorbar_ticks="outside")
        plot = json.loads(fig.to_json())

    return {
        "name": hist_name,
        "type": "heatmap",
        "plot": plot["data"],
        "layout": plot["layout"],
    }


def _prune(values, last_n=0, skip_first_n=0, skip_last_n=0):
    """inline function to select first or last items of input list

    :param values: input list to select from
    :param int last_n: select last 'n' items of values. default is 0.
    :param int skip_first_n: skip first n items of values. default is 0. last_n takes precedence.
    :param int skip_last_n: in plot skip last 'n' periods. last_n takes precedence (optional)
    :return: list of selected values
    """
    if last_n > 0:
        return values[-last_n:]
    if skip_first_n > 0:
        values = values[skip_first_n:]
    if skip_last_n > 0:
        values = values[:-skip_last_n]
    return values



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


"""Hartigan & Hartigan's dip test for unimodality

How to enable this extension:
    - Install te diptest package: `pip install diptest` or `pip install popmon[diptest]`
    - To show the diptest values in your report:
        settings.report.show_stats.append("diptest*")
        OR
        settings.report.extended_report = True

"""
import numpy as np

from popmon.analysis import Profiles
from popmon.extensions.extension import Extension


def extension():
    from diptest import diptest

    @Profiles.register(
        key=["diptest_value", "diptest_pvalue"],
        description=[
            "diptest value for Hartigan & Hartigan's test for unimodality",
            "p-value for the diptest",
        ],
        dim=1,
        htype="num",
    )
    def diptest_profile(bin_centers, bin_values, bin_width, rng=None):
        if rng is None:
            rng = np.random.default_rng()

        counts = bin_values.astype(int)
        n = counts.sum()
        hbw = bin_width / 2

        # unpack histogram into ordered samples
        sample = np.repeat(bin_centers, counts)

        # uniform noise
        sample_noise = sample + rng.uniform(-hbw, hbw, n)

        # compute diptest
        dip, pval = diptest(sample_noise)
        return dip, pval


class Diptest(Extension):
    name = "diptest"
    requirements = ["diptest"]
    extension = extension



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import numpy as np
import pandas as pd

from ...base import Module
from ...hist.hist_utils import get_bin_centers, is_numeric, is_timestamp


class HistProfiler(Module):
    """Generate profiles of histograms using default statistical functions.

    Profiles are:

    - 1 dim histograms, all: 'count', 'filled', 'distinct', 'nan', 'most_probable_value', 'overflow', 'underflow'.
    - 1 dim histograms, numeric: mean, std, min, max, p01, p05, p16, p50, p84, p95, p99.
    - 1 dim histograms, boolean: fraction of true entries.
    - 2 dim histograms: count, phi_k correlation constant, p-value and Z-score of contingency test.
    - n dim histograms: count (n >= 3)

    :param str read_key: key of the input test data to read from the datastore
    :param str store_key: key of the output data to store in the datastore
    :param list features: features of data-frames to pick up from input data (optional)
    :param list ignore_features: features to ignore (optional)
    :param list var_timestamp: list of timestamp variables (optional)
    :param str hist_col: key for histogram in split dictionary
    :param str index_col: key for index in split dictionary
    :param dict stats_functions: function_name, function(bin_labels, bin_counts) dictionary
    """

    _input_keys = ("read_key",)
    _output_keys = ("store_key",)

    def __init__(
        self,
        read_key,
        store_key,
        features=None,
        ignore_features=None,
        var_timestamp=None,
        hist_col="histogram",
        index_col="date",
        stats_functions=None,
    ):
        super().__init__()
        self.read_key = read_key
        self.store_key = store_key

        self.features = features or []
        self.ignore_features = ignore_features or []
        self.var_timestamp = var_timestamp or []
        self.hist_col = hist_col
        self.index_col = index_col

        if stats_functions is not None:
            raise NotImplementedError()

    def _profile_1d_histogram(self, name, hist):
        from popmon.analysis import Profiles

        # preprocessing value counts and TS
        is_num = is_numeric(hist)
        is_ts = is_timestamp(hist) or name in self.var_timestamp

        bin_labels, values = get_bin_centers(hist)
        bin_counts = np.array([v.entries for v in values])

        if len(bin_counts) == 0:
            self.logger.warning(f'Histogram "{name}" is empty; skipping.')
            return {}

        if is_ts:
            to_timestamp = np.vectorize(lambda x: pd.to_datetime(x).value)
            bin_labels = to_timestamp(bin_labels)

        otype = "num" if is_num else "cat"

        # calc 1d-histogram statistics
        profile = {}
        if otype == "cat":
            args = [bin_labels, bin_counts]
        else:
            bin_width = hist.bin_width()
            args = [bin_labels, bin_counts, bin_width]

        profile.update(Profiles.run(args, dim=1, htype=otype))

        args = [bin_labels, bin_counts]
        profile.update(Profiles.run(args, dim=1, htype="all"))

        # difference between htype=None and htype="all" are arguments (bin labels vs hist)
        profile.update(Profiles.run([hist], dim=1, htype=None))
        profile.update(Profiles.run([hist], dim=-1, htype=None))

        # postprocessing TS
        if is_ts:
            profile = {
                k: pd.Timestamp(v) if k != "std" else pd.Timedelta(v)
                for k, v in profile.items()
            }

        return profile

    def _profile_nd_histogram(self, name, hist, dim):
        from popmon.analysis import Profiles

        if hist.n_dim < dim:
            self.logger.warning(
                f"Histogram {name} has {hist.n_dim} dimensions (<{dim}); cannot profile. Returning empty."
            )
            return {}

        # calc nd-histogram statistics
        profile = Profiles.run([hist], dim=dim, htype=None)
        profile.update(Profiles.run([hist], dim=dim, htype="all"))
        profile.update(Profiles.run([hist], dim=dim, htype="num"))
        profile.update(Profiles.run([hist], dim=dim, htype="cat"))

        profile.update(Profiles.run([hist], dim=-1, htype=None))
        return profile

    def _profile_hist(self, split, hist_name):
        from popmon.analysis.profiling import Profiles

        if len(split) == 0:
            self.logger.error(f'Split histograms dict "{hist_name}" is empty. Return.')
            return []

        hist0 = split[0][self.hist_col]
        dimension = hist0.n_dim
        is_num = is_numeric(hist0)
        htype = "num" if is_num else "cat"

        # these are the profiled quantities we will monitor
        expected_fields = (
            Profiles.get_keys_by_dim_and_htype(dim=dimension, htype=htype)
            + Profiles.get_keys_by_dim_and_htype(dim=dimension, htype="all")
            + Profiles.get_keys_by_dim_and_htype(dim=dimension, htype=None)
        )

        # profiles regardless of dim and htype (e.g. count)
        expected_fields += Profiles.get_keys_by_dim_and_htype(dim=None, htype=None)

        # profiles regardless of dim
        expected_fields += Profiles.get_keys_by_dim_and_htype(dim=-1, htype=htype)
        expected_fields += Profiles.get_keys_by_dim_and_htype(dim=-1, htype="all")
        expected_fields += Profiles.get_keys_by_dim_and_htype(dim=-1, htype=None)

        expected_fields += [self.index_col, self.hist_col]

        # now loop over split-axis, e.g. time index, and profile each sub-hist x:y
        profile_list = []
        for hist_dict in split:
            index, hist = hist_dict[self.index_col], hist_dict[self.hist_col]

            profile = {self.index_col: index, self.hist_col: hist}

            if dimension == 1:
                profile.update(self._profile_1d_histogram(hist_name, hist))
            else:
                profile.update(
                    self._profile_nd_histogram(hist_name, hist, dim=dimension)
                )

            if sorted(profile.keys()) != sorted(expected_fields):
                self.logger.error(
                    f'Could not extract full profile for sub-hist "{hist_name} {index}".'
                    f"Differences: {set(profile.keys()).symmetric_difference(set(expected_fields))}. Skipping."
                )
            else:
                profile_list.append(profile)

        return profile_list

    def transform(self, data: dict) -> dict:
        self.logger.info(
            f'Profiling histograms "{self.read_key}" as "{self.store_key}"'
        )
        features = self.get_features(list(data.keys()))

        profiled = {}
        for feature in features[:]:
            df = self.get_datastore_object(data, feature, dtype=pd.DataFrame)
            hist_split_list = df.reset_index().to_dict("records")

            self.logger.debug(f'Profiling histogram "{feature}".')
            profile_list = self._profile_hist(split=hist_split_list, hist_name=feature)
            if len(profile_list) > 0:
                profiled[feature] = pd.DataFrame(profile_list).set_index(
                    [self.index_col]
                )

        return profiled



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


# histogram and report functions
from histogrammar.dfinterface.make_histograms import (
    get_bin_specs,
    get_time_axes,
    make_histograms,
)

# pandas/spark dataframe decorators
from popmon import decorators

from .config import Settings
from .extensions import extensions
from .pipeline.metrics import df_stability_metrics, stability_metrics
from .pipeline.report import df_stability_report, stability_report
from .stitching import stitch_histograms
from .version import version as __version__

__all__ = [
    "get_bin_specs",
    "get_time_axes",
    "make_histograms",
    "decorators",
    "df_stability_metrics",
    "stability_metrics",
    "df_stability_report",
    "stability_report",
    "stitch_histograms",
    "__version__",
    "Settings",
    "extensions",
]



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import fnmatch
from textwrap import shorten
from typing import Callable, List, Optional

import pandas as pd
from joblib import Parallel, delayed


def short_date(date):
    """
    Shorten date string to length of 22
    """
    if isinstance(date, pd.Timestamp):
        # Drop the time of day when midnight or noon
        if date.hour in [0, 12] and date.minute == 0 and date.second == 0:
            d = str(date).split(" ")[0]
        else:
            d = str(date)
    else:
        d = str(date)

    return shorten(d, width=22, placeholder="")


def filter_metrics(
    metrics, ignore_stat_endswith: List[str], show_stats: Optional[List[str]] = None
):
    """
    Filter metrics by excluding based on suffix and/or including when matching a pattern
    """
    metrics = [
        m for m in metrics if not any(m.endswith(s) for s in ignore_stat_endswith)
    ]
    if show_stats is not None:
        metrics = [
            m
            for m in metrics
            if any(fnmatch.fnmatch(m, pattern) for pattern in show_stats)
        ]
    return metrics


def parallel(func: Callable, args_list, mode="args"):
    """
    Routine for parallel processing
    """
    from popmon.config import parallel_args

    if parallel_args["n_jobs"] == 1:
        results = [
            func(*args) if mode == "args" else func(**args) for args in args_list
        ]
    else:
        results = Parallel(**parallel_args)(
            delayed(func)(*args) if mode == "args" else delayed(func)(**args)
            for args in args_list
        )
    return results



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import importlib.util
from typing import Callable, List


def is_installed(package):
    is_present = importlib.util.find_spec(package)
    return is_present is not None


class Extension:
    name: str
    requirements: List[str]
    extension: Callable

    @property
    def extras(self):
        return {self.name: self.requirements}

    def check(self):
        if all(is_installed(package) for package in self.requirements):
            func = self.extension
            func = func.__func__
            func()



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import fnmatch
from typing import Optional

import numpy as np
import pandas as pd

from ..base import Module


class AlertsSummary(Module):
    """The module AlertsSummary combines the alerts-summaries of all individual features

    It combines the alerts-summaries of all individual features into an artificial feature "_AGGREGATE_".
    """

    _input_keys = ("read_key",)
    _output_keys = ("store_key",)

    def __init__(
        self,
        read_key,
        store_key="",
        features=None,
        ignore_features=None,
        combined_variable="_AGGREGATE_",
    ):
        """Initialize an instance of AlertsSummary module.

        :param str read_key: key of input data to read from datastore.
        :param str store_key: key of output data to store in datastore (optional).
        :param str combined_variable: name of artificial variable that combines all alerts. default is '_AGGREGATE_'.
        :param list features: features of data frames to pick up from input data (optional).
        :param list ignore_features: list of features to ignore (optional).
        """
        super().__init__()
        self.read_key = read_key
        self.store_key = store_key or self.read_key
        self.features = features or []
        self.ignore_features = ignore_features or []
        self.combined_variable = combined_variable

    def transform(self, data: dict) -> Optional[dict]:
        # determine all possible features, used for the comparison below
        features = self.get_features(list(data.keys()))
        if len(features) == 0:
            return None

        self.logger.info(
            f'Combining alerts into artificial variable "{self.combined_variable}"'
        )

        # STEP 1: loop over features where alerts exist
        df_list = []
        for feature in features:
            # basic checks if feature object is filled correctly
            df = (self.get_datastore_object(data, feature, dtype=pd.DataFrame)).copy(
                deep=False
            )
            df.columns = [f"{feature}_{c}" for c in df.columns]
            df_list.append(df)

        # the different features could technically have different indices.
        # will only merge alerts if all indices are the same
        if len(df_list) >= 2:
            for df in df_list[1:]:
                if not np.array_equal(df_list[0].index, df.index):
                    self.logger.warning(
                        "indices of features are different. no alerts summary generated."
                    )
                    return None

        # STEP 2: Concatenate the dataframes, there was one for each original feature.
        tlv = pd.concat(df_list, axis=1)
        dfc = pd.DataFrame(index=tlv.index)

        # colors of traffic lights
        for color in ["green", "yellow", "red"]:
            cols = fnmatch.filter(tlv.columns, f"*_n_{color}")
            dfc[f"n_{color}"] = tlv[cols].values.sum(axis=1) if len(cols) else 0

        # store combination of traffic alerts
        data[self.combined_variable] = dfc
        return data



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import logging

from histogrammar.dfinterface.make_histograms import get_bin_specs, make_histograms

from ..config import Settings
from ..pipeline.metrics_pipelines import create_metrics_pipeline

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s %(levelname)s [%(module)s]: %(message)s"
)
logger = logging.getLogger()


def stability_metrics(
    hists,
    settings: Settings,
    reference=None,
):
    """Create a data stability monitoring datastore for given dict of input histograms.

    :param dict hists: input histograms to be profiled and monitored over time.
    :param popmon.config.Settings settings: popmon configuration object
    :param reference: histograms used as reference. default is None
    :return: dict with results of metrics pipeline
    """

    if not isinstance(hists, dict):
        raise TypeError("hists should be a dict of histogrammar histograms.")

    if isinstance(settings.time_axis, str) and len(settings.time_axis) == 0:
        settings._set_time_axis_hists(hists)

    kwargs = {}
    if (
        settings.reference_type in ["external", "self_split"]
        and "ref_hists_key" not in kwargs
    ):
        kwargs["ref_hists_key"] = "ref_hists"

    pipeline = create_metrics_pipeline(
        settings=settings,
        reference=reference,
        hists_key="hists",
        **kwargs,
    )

    datastore = {"hists": hists}
    if settings.reference_type in ["external", "self_split"]:
        datastore["ref_hists"] = reference

    return pipeline.transform(datastore)


def df_stability_metrics(
    df,
    settings: Settings = None,
    time_width=None,
    time_offset=0,
    var_dtype=None,
    reference=None,
    **kwargs,
):
    """Create a data stability monitoring html datastore for given pandas or spark dataframe.

    :param df: input pandas/spark dataframe to be profiled and monitored over time.
    :param popmon.config.Settings settings: popmon configuration object
    :param time_width: bin width of time axis. str or number (ns). note: bin_specs takes precedence. (optional)

        .. code-block:: text

            Examples: '1w', 3600e9 (number of ns),
                      anything understood by pd.Timedelta(time_width).value

    :param time_offset: bin offset of time axis. str or number (ns). note: bin_specs takes precedence. (optional)

        .. code-block:: text

            Examples: '1-1-2020', 0 (number of ns since 1-1-1970),
                      anything parsed by pd.Timestamp(time_offset).value

    :param dict var_dtype: dictionary with specified datatype per feature. auto-guessed when not provided.
    :param reference: reference dataframe or histograms. default is None
    :param kwargs: residual keyword arguments, passed on to stability_report()
    :return: dict with results of metrics pipeline
    """
    if settings is None:
        settings = Settings(**kwargs)

    if len(settings.time_axis) == 0:
        settings._set_time_axis_dataframe(df)
        logger.info(f'Time-axis automatically set to "{settings.time_axis}"')

    if settings.time_axis not in df.columns:
        raise ValueError(
            f'time_axis  "{settings.time_axis}" not found in columns of dataframe.'
        )

    if (
        reference is not None
        and not isinstance(reference, dict)
        and settings.time_axis not in reference.columns
    ):
        raise ValueError(
            f'time_axis  "{settings.time_axis}" not found in columns of reference dataframe.'
        )

    if settings.features is not None:
        # by now time_axis is defined. ensure that all histograms start with it.
        settings._ensure_features_time_axis()

    # interpret time_width and time_offset
    if time_width is not None:
        if not isinstance(time_width, (str, int, float)):
            raise TypeError
        if not isinstance(time_offset, (str, int, float)):
            raise TypeError

        settings._set_bin_specs_by_time_width_and_offset(time_width, time_offset)

    if reference is not None:
        if settings.reference_type not in ["external", "self_split"]:
            raise TypeError(
                "When providing a `reference`, the `reference_type` should be equal to `external` or `self_split`"
            )

        if isinstance(reference, dict):
            # 1. reference is dict of histograms
            # extract features and bin_specs from reference histograms
            reference_hists = reference
            if settings.features is not None or settings.bin_specs != {}:
                raise ValueError(
                    "When providing a reference, the `features` and `bin_specs` settings should be default (as they are overriden)"
                )

            settings.features = list(reference_hists.keys())
            settings.bin_specs = get_bin_specs(reference_hists)
        else:
            # 2. reference is pandas or spark dataframe
            # generate histograms and return updated features, bin_specs, time_axis, etc.
            (
                reference_hists,
                settings.features,
                settings.bin_specs,
                settings.time_axis,
                var_dtype,
            ) = make_histograms(
                reference,
                settings.features,
                settings.binning,
                settings.bin_specs,
                settings.time_axis,
                var_dtype,
                ret_specs=True,
            )
        kwargs["reference_hists"] = reference_hists

    # use the same features, bin_specs, time_axis, etc as for reference hists
    hists = make_histograms(
        df,
        features=settings.features,
        binning=settings.binning,
        bin_specs=settings.bin_specs,
        time_axis=settings.time_axis,
        var_dtype=var_dtype,
    )

    # generate data stability report
    return stability_metrics(
        hists,
        settings=settings,
        reference=reference,
    )



import json

from setuptools import find_packages, setup

__version__ = "1.3.0"

with open("requirements.txt") as f:
    REQUIREMENTS = f.read().splitlines()

# read the contents of abstract file
with open("README.rst", encoding="utf-8") as f:
    long_description = f.read()

# read dynamically generated extras from json file
with open("extras.json") as f:
    EXTRAS = json.loads(f.read())


def setup_package() -> None:
    """The main setup method.

    It is responsible for setting up and installing the package.
    """
    setup(
        name="popmon",
        version=__version__,
        url="https://github.com/ing-bank/popmon",
        license="MIT",
        author="ING Wholesale Banking Advanced Analytics",
        description="Monitor the stability of a pandas or spark dataset",
        keywords="pandas spark data-science data-analysis monitoring statistics python jupyter ipython",
        long_description=long_description,
        long_description_content_type="text/x-rst",
        python_requires=">=3.6",
        packages=find_packages(),
        install_requires=REQUIREMENTS,
        extras_require=EXTRAS,
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        # files to be shipped with the installation, under: popmon/popmon/
        # after installation, these can be found with the functions in resources.py
        package_data={
            "popmon": [
                "visualization/templates/*.html",
                "visualization/templates/assets/css/*.css",
                "visualization/templates/assets/js/*.js",
                "test_data/*.csv.gz",
                "test_data/*.json*",
                "notebooks/popmon*tutorial*.ipynb",
            ]
        },
        entry_points={
            "console_scripts": ["popmon_run = popmon.pipeline.amazing_pipeline:run"]
        },
    )


if __name__ == "__main__":
    setup_package()
